# Copyright 2021-2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import math
import os.path
import re
import tempfile
import unittest
import urllib.request
from typing import Any, Dict, cast

import jsonschema
import numpy as np
import pandas as pd
import sklearn
from category_encoders.hashing import HashingEncoder as SkHashingEncoder
from sklearn.feature_selection import SelectKBest as SkSelectKBest
from sklearn.impute import SimpleImputer as SkSimpleImputer
from sklearn.metrics import accuracy_score as sk_accuracy_score
from sklearn.metrics import make_scorer
from sklearn.metrics import r2_score as sk_r2_score
from sklearn.model_selection import KFold
from sklearn.model_selection import cross_val_score as sk_cross_val_score
from sklearn.model_selection import cross_validate as sk_cross_validate
from sklearn.pipeline import make_pipeline as sk_make_pipeline
from sklearn.preprocessing import MinMaxScaler as SkMinMaxScaler
from sklearn.preprocessing import OneHotEncoder as SkOneHotEncoder
from sklearn.preprocessing import OrdinalEncoder as SkOrdinalEncoder
from sklearn.preprocessing import StandardScaler as SkStandardScaler

import lale.datasets
import lale.datasets.openml
import lale.lib.aif360
import lale.type_checking
from lale.datasets import pandas2spark
from lale.datasets.data_schemas import (
    SparkDataFrameWithIndex,
    add_table_name,
    forward_metadata,
    get_index_name,
)
from lale.datasets.multitable.fetch_datasets import fetch_go_sales_dataset
from lale.expressions import it
from lale.helpers import _ensure_pandas, create_data_loader
from lale.lib.rasl import BatchedBaggingClassifier, ConcatFeatures, Convert
from lale.lib.rasl import HashingEncoder as RaslHashingEncoder
from lale.lib.rasl import Map
from lale.lib.rasl import MinMaxScaler as RaslMinMaxScaler
from lale.lib.rasl import OneHotEncoder as RaslOneHotEncoder
from lale.lib.rasl import OrdinalEncoder as RaslOrdinalEncoder
from lale.lib.rasl import PrioBatch, PrioStep, Project, Scan
from lale.lib.rasl import SelectKBest as RaslSelectKBest
from lale.lib.rasl import SimpleImputer as RaslSimpleImputer
from lale.lib.rasl import StandardScaler as RaslStandardScaler
from lale.lib.rasl import accuracy_score as rasl_accuracy_score
from lale.lib.rasl import categorical
from lale.lib.rasl import cross_val_score as rasl_cross_val_score
from lale.lib.rasl import cross_validate as rasl_cross_validate
from lale.lib.rasl import csv_data_loader, fit_with_batches
from lale.lib.rasl import get_scorer as rasl_get_scorer
from lale.lib.rasl import mockup_data_loader, openml_data_loader
from lale.lib.rasl import r2_score as rasl_r2_score
from lale.lib.sklearn import (
    DecisionTreeClassifier,
    LogisticRegression,
    RandomForestClassifier,
    SGDClassifier,
)
from lale.operators import TrainedPipeline

assert sklearn.__version__ >= "1.0", sklearn.__version__


class TestDatasets(unittest.TestCase):
    def test_openml_creditg_arff(self):
        batches = openml_data_loader("credit-g", 340)
        n_rows_found = 0
        n_batches_found = 0
        for bX, by in batches:
            n_batches_found += 1
            n_rows_batch, n_columns_batch = bX.shape
            n_rows_found += n_rows_batch
            self.assertEqual(n_rows_batch, len(by))
            self.assertEqual(n_columns_batch, 20)
        self.assertEqual(n_batches_found, 3)
        self.assertEqual(n_rows_found, 1000)

    def test_autoai_creditg_csv(self):
        with tempfile.TemporaryDirectory() as tmpdir_name:
            url = "https://raw.githubusercontent.com/pmservice/wml-sample-models/master/autoai/credit-risk-prediction/data/german_credit_data_biased_training.csv"
            file_name = os.path.join(tmpdir_name, "credit-g.csv")
            urllib.request.urlretrieve(url, file_name)
            assert os.path.exists(file_name)
            n_rows = 5000
            n_batches = 3
            rows_per_batch = (n_rows + n_batches - 1) // n_batches
            batches = csv_data_loader(file_name, "Risk", rows_per_batch)
            n_rows_found = 0
            n_batches_found = 0
            for bX, by in batches:
                n_batches_found += 1
                n_rows_batch, n_columns_batch = bX.shape
                n_rows_found += n_rows_batch
                self.assertEqual(n_rows_batch, len(by))
                self.assertEqual(n_columns_batch, 20)
            self.assertEqual(n_batches_found, n_batches)
            self.assertEqual(n_rows_found, n_rows)


def _check_trained_min_max_scaler(test, op1, op2, msg):
    test.assertEqual(list(op1.data_min_), list(op2.data_min_), msg)
    test.assertEqual(list(op1.data_max_), list(op2.data_max_), msg)
    test.assertEqual(list(op1.data_range_), list(op2.data_range_), msg)
    test.assertEqual(list(op1.scale_), list(op2.scale_), msg)
    test.assertEqual(list(op1.min_), list(op2.min_), msg)
    test.assertEqual(op1.n_features_in_, op2.n_features_in_, msg)
    test.assertEqual(op1.n_samples_seen_, op2.n_samples_seen_, msg)


class TestMinMaxScaler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: fetch_go_sales_dataset(tgt) for tgt in targets}

    def test_get_params(self):
        sk_scaler = SkMinMaxScaler()
        rasl_scaler = RaslMinMaxScaler()
        sk_params = sk_scaler.get_params()
        rasl_params = rasl_scaler.get_params()
        self.assertDictContainsSubset(sk_params, rasl_params)

    def test_error(self):
        with self.assertRaisesRegex(
            jsonschema.ValidationError,
            re.compile(r"MinMaxScaler\(copy=False\)", re.MULTILINE | re.DOTALL),
        ):
            _ = RaslMinMaxScaler(copy=False)
        with self.assertRaisesRegex(
            jsonschema.ValidationError,
            re.compile(r"MinMaxScaler\(clip=True\)", re.MULTILINE | re.DOTALL),
        ):
            _ = RaslMinMaxScaler(clip=True)

    def test_fit(self):
        columns = ["Product number", "Quantity", "Retailer code"]
        pandas_data = self.tgt2datasets["pandas"][0][columns]
        sk_scaler = SkMinMaxScaler()
        sk_trained = sk_scaler.fit(pandas_data)
        rasl_scaler = RaslMinMaxScaler()
        for tgt, go_sales in self.tgt2datasets.items():
            data = go_sales[0][columns]
            if tgt == "spark-with-index":
                data = SparkDataFrameWithIndex(data)
            rasl_trained = rasl_scaler.fit(data)
            _check_trained_min_max_scaler(self, sk_trained, rasl_trained.impl, "pandas")

    def test_transform(self):
        columns = ["Product number", "Quantity", "Retailer code"]
        pandas_data = self.tgt2datasets["pandas"][0][columns]
        sk_scaler = SkMinMaxScaler()
        sk_trained = sk_scaler.fit(pandas_data)
        sk_transformed = sk_trained.transform(pandas_data)
        rasl_scaler = RaslMinMaxScaler()
        for tgt, go_sales in self.tgt2datasets.items():
            data = go_sales[0][columns]
            if tgt == "spark-with-index":
                data = SparkDataFrameWithIndex(data)
            rasl_trained = rasl_scaler.fit(data)
            rasl_transformed = rasl_trained.transform(data)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(rasl_transformed), "index")
            rasl_transformed = _ensure_pandas(rasl_transformed)
            self.assertAlmostEqual(sk_transformed[0, 0], rasl_transformed.iloc[0, 0])
            self.assertAlmostEqual(sk_transformed[0, 1], rasl_transformed.iloc[0, 1])
            self.assertAlmostEqual(sk_transformed[0, 2], rasl_transformed.iloc[0, 2])
            self.assertAlmostEqual(sk_transformed[10, 0], rasl_transformed.iloc[10, 0])
            self.assertAlmostEqual(sk_transformed[10, 1], rasl_transformed.iloc[10, 1])
            self.assertAlmostEqual(sk_transformed[10, 2], rasl_transformed.iloc[10, 2])
            self.assertAlmostEqual(sk_transformed[20, 0], rasl_transformed.iloc[20, 0])
            self.assertAlmostEqual(sk_transformed[20, 1], rasl_transformed.iloc[20, 1])
            self.assertAlmostEqual(sk_transformed[20, 2], rasl_transformed.iloc[20, 2])

    def test_fit_range(self):
        columns = ["Product number", "Quantity", "Retailer code"]
        pandas_data = self.tgt2datasets["pandas"][0][columns]
        sk_scaler = SkMinMaxScaler(feature_range=(-5, 5))
        sk_trained = sk_scaler.fit(pandas_data)
        for tgt, go_sales in self.tgt2datasets.items():
            data = go_sales[0][columns]
            if tgt == "spark-with-index":
                data = SparkDataFrameWithIndex(data)
            rasl_scaler = RaslMinMaxScaler(feature_range=(-5, 5))
            rasl_trained = rasl_scaler.fit(data)
            _check_trained_min_max_scaler(self, sk_trained, rasl_trained.impl, "pandas")

    def test_transform_range(self):
        columns = ["Product number", "Quantity", "Retailer code"]
        pandas_data = self.tgt2datasets["pandas"][0][columns]
        sk_scaler = SkMinMaxScaler(feature_range=(-5, 5))
        sk_trained = sk_scaler.fit(pandas_data)
        sk_transformed = sk_trained.transform(pandas_data)
        rasl_scaler = RaslMinMaxScaler(feature_range=(-5, 5))
        for tgt, go_sales in self.tgt2datasets.items():
            data = go_sales[0][columns]
            if tgt == "spark-with-index":
                data = SparkDataFrameWithIndex(data)
            rasl_trained = rasl_scaler.fit(data)
            rasl_transformed = rasl_trained.transform(data)
            rasl_transformed = _ensure_pandas(rasl_transformed)
            self.assertAlmostEqual(sk_transformed[0, 0], rasl_transformed.iloc[0, 0])
            self.assertAlmostEqual(sk_transformed[0, 1], rasl_transformed.iloc[0, 1])
            self.assertAlmostEqual(sk_transformed[0, 2], rasl_transformed.iloc[0, 2])
            self.assertAlmostEqual(sk_transformed[10, 0], rasl_transformed.iloc[10, 0])
            self.assertAlmostEqual(sk_transformed[10, 1], rasl_transformed.iloc[10, 1])
            self.assertAlmostEqual(sk_transformed[10, 2], rasl_transformed.iloc[10, 2])
            self.assertAlmostEqual(sk_transformed[20, 0], rasl_transformed.iloc[20, 0])
            self.assertAlmostEqual(sk_transformed[20, 1], rasl_transformed.iloc[20, 1])
            self.assertAlmostEqual(sk_transformed[20, 2], rasl_transformed.iloc[20, 2])

    def test_partial_fit(self):
        columns = ["Product number", "Quantity", "Retailer code"]
        data = self.tgt2datasets["pandas"][0][columns]
        for tgt in self.tgt2datasets.keys():
            sk_scaler = SkMinMaxScaler()
            rasl_scaler = RaslMinMaxScaler()
            for lower, upper in [[0, 10], [10, 100], [100, data.shape[0]]]:
                data_so_far = data[0:upper]
                data_delta = data[lower:upper]
                if tgt == "pandas":
                    pass
                elif tgt == "spark":
                    data_delta = pandas2spark(data_delta)
                elif tgt == "spark-with-index":
                    data_delta = pandas2spark(data_delta, with_index=True)
                else:
                    assert False
                sk_trained = sk_scaler.fit(data_so_far)
                rasl_trained = rasl_scaler.partial_fit(data_delta)
                _check_trained_min_max_scaler(self, sk_trained, rasl_trained.impl, tgt)


class TestPipeline(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: {} for tgt in targets}

        def add_df(name, df):
            cls.tgt2datasets["pandas"][name] = df
            cls.tgt2datasets["spark"][name] = pandas2spark(df)
            cls.tgt2datasets["spark-with-index"][name] = pandas2spark(
                df, with_index=True
            )

        X, y = load_iris(as_frame=True, return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(X, y)
        add_df("X_train", X_train)
        add_df("X_test", X_test)
        add_df("y_train", y_train)
        add_df("y_test", y_test)

    def test_pipeline(self):
        for tgt, datasets in self.tgt2datasets.items():
            X_train, X_test = (datasets["X_train"], datasets["X_test"])
            y_train = self.tgt2datasets["pandas"]["y_train"]
            pipeline = (
                RaslMinMaxScaler() >> Convert(astype="pandas") >> LogisticRegression()
            )
            trained = pipeline.fit(X_train, y_train)
            _ = trained.predict(X_test)


def _check_trained_ordinal_encoder(test, op1, op2, msg):
    test.assertEqual(list(op1.feature_names_in_), list(op2.feature_names_in_), msg)
    test.assertEqual(len(op1.categories_), len(op2.categories_), msg)
    for i in range(len(op1.categories_)):
        test.assertEqual(list(op1.categories_[i]), list(op2.categories_[i]), msg)


class TestSelectKBest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from sklearn.datasets import load_digits

        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: {} for tgt in targets}

        def add_df(name, df):
            cls.tgt2datasets["pandas"][name] = df
            cls.tgt2datasets["spark"][name] = pandas2spark(df)
            cls.tgt2datasets["spark-with-index"][name] = pandas2spark(
                df, with_index=True
            )

        X, y = load_digits(return_X_y=True, as_frame=True)
        X = add_table_name(X, "X")
        y = add_table_name(y, "y")
        add_df("X", X)
        add_df("y", y)

    def _check_trained(self, sk_trained, rasl_trained, msg=""):
        for i in range(len(sk_trained.scores_)):
            if not (
                np.isnan(sk_trained.scores_[i])
                and np.isnan(rasl_trained.impl.scores_[i])
            ):
                self.assertAlmostEqual(
                    sk_trained.scores_[i],
                    rasl_trained.impl.scores_[i],
                    msg=f"{msg}: {i}",
                )
            if not (
                np.isnan(sk_trained.pvalues_[i])
                and np.isnan(rasl_trained.impl.pvalues_[i])
            ):
                self.assertAlmostEqual(
                    sk_trained.pvalues_[i],
                    rasl_trained.impl.pvalues_[i],
                    msg=f"{msg}: {i}",
                )
        self.assertEqual(
            sk_trained.n_features_in_, rasl_trained.impl.n_features_in_, msg
        )

    def test_fit(self):
        sk_trainable = SkSelectKBest(k=20)
        X, y = self.tgt2datasets["pandas"]["X"], self.tgt2datasets["pandas"]["y"]
        sk_trained = sk_trainable.fit(X, y)
        rasl_trainable = RaslSelectKBest(k=20)
        for tgt, datasets in self.tgt2datasets.items():
            X, y = datasets["X"], datasets["y"]
            if tgt == "spark":
                with self.assertRaises(ValueError):
                    rasl_trained = rasl_trainable.fit(X, y)
            else:
                rasl_trained = rasl_trainable.fit(X, y)
                self._check_trained(sk_trained, rasl_trained, tgt)

    def test_transform(self):
        sk_trainable = SkSelectKBest(k=20)
        X, y = self.tgt2datasets["pandas"]["X"], self.tgt2datasets["pandas"]["y"]
        sk_trained = sk_trainable.fit(X, y)
        sk_transformed = sk_trained.transform(X)
        rasl_trainable = RaslSelectKBest(k=20)
        for tgt, datasets in self.tgt2datasets.items():
            X, y = datasets["X"], datasets["y"]
            if tgt == "spark":
                with self.assertRaises(ValueError):
                    rasl_trained = rasl_trainable.fit(X, y)
                continue
            rasl_trained = rasl_trainable.fit(X, y)
            rasl_transformed = rasl_trained.transform(X)
            self._check_trained(sk_trained, rasl_trained, tgt)
            rasl_transformed = _ensure_pandas(rasl_transformed)
            self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
            for row_idx in range(sk_transformed.shape[0]):
                for col_idx in range(sk_transformed.shape[1]):
                    self.assertAlmostEqual(
                        sk_transformed[row_idx, col_idx],
                        rasl_transformed.iloc[row_idx, col_idx],
                        msg=(row_idx, col_idx),
                    )

    def test_partial_fit(self):
        rasl_trainable = RaslSelectKBest(k=20)
        X, y = self.tgt2datasets["pandas"]["X"], self.tgt2datasets["pandas"]["y"]
        for lower, upper in [[0, 100], [100, 200], [200, X.shape[0]]]:
            X_so_far, y_so_far = X[0:upper], y[0:upper]
            sk_trainable = SkSelectKBest(k=20)
            sk_trained = sk_trainable.fit(X_so_far, y_so_far)
            X_delta, y_delta = X[lower:upper], y[lower:upper]
            rasl_trained = rasl_trainable.partial_fit(X_delta, y_delta)
            self._check_trained(
                sk_trained, rasl_trained, f"lower: {lower}, upper: {upper}"
            )


class TestOrdinalEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2gosales = {tgt: fetch_go_sales_dataset(tgt) for tgt in targets}
        cls.tgt2creditg = {
            tgt: lale.datasets.openml.fetch(
                "credit-g",
                "classification",
                preprocess=False,
                astype=tgt,
            )
            for tgt in targets
        }

    def _check_last_trained(self, op1, op2, msg):
        _check_trained_ordinal_encoder(
            self, op1.get_last().impl, op2.get_last().impl, msg
        )

    def test_fit(self):
        prefix = Scan(table=it.go_daily_sales) >> Map(
            columns={"retailer": it["Retailer code"], "method": it["Order method code"]}
        )
        encoder_args = {"handle_unknown": "use_encoded_value", "unknown_value": np.nan}
        rasl_trainable = prefix >> RaslOrdinalEncoder(**encoder_args)
        sk_trainable = prefix >> SkOrdinalEncoder(**encoder_args)
        sk_trained = sk_trainable.fit(self.tgt2gosales["pandas"])
        for tgt, datasets in self.tgt2gosales.items():
            rasl_trained = rasl_trainable.fit(datasets)
            self._check_last_trained(sk_trained, rasl_trained, tgt)

    def test_partial_fit(self):
        prefix = Scan(table=it.go_daily_sales) >> Map(
            columns={"retailer": it["Retailer code"], "method": it["Order method code"]}
        )
        pandas_data = prefix.transform(self.tgt2gosales["pandas"])
        encoder_args = {"handle_unknown": "use_encoded_value", "unknown_value": np.nan}
        for tgt in self.tgt2gosales.keys():
            rasl_op = RaslOrdinalEncoder(**encoder_args)
            for lower, upper in [[0, 10], [10, 100], [100, pandas_data.shape[0]]]:
                data_so_far = pandas_data[0:upper]
                sk_op = SkOrdinalEncoder(**encoder_args).fit(data_so_far)
                data_delta = pandas_data[lower:upper]
                if tgt == "pandas":
                    pass
                elif tgt == "spark":
                    data_delta = pandas2spark(data_delta)
                elif tgt == "spark-with-index":
                    data_delta = pandas2spark(data_delta, with_index=True)
                else:
                    assert False
                rasl_op = rasl_op.partial_fit(data_delta)
                _check_trained_ordinal_encoder(
                    self,
                    sk_op,
                    rasl_op.impl,
                    f"tgt {tgt}, lower {lower}, upper {upper}",
                )

    def test_transform(self):
        prefix = Scan(table=it.go_daily_sales) >> Map(
            columns={"retailer": it["Retailer code"], "method": it["Order method code"]}
        )
        encoder_args = {"handle_unknown": "use_encoded_value", "unknown_value": np.nan}
        rasl_trainable = prefix >> RaslOrdinalEncoder(**encoder_args)
        sk_trainable = prefix >> SkOrdinalEncoder(**encoder_args)
        sk_trained = sk_trainable.fit(self.tgt2gosales["pandas"])
        sk_transformed = sk_trained.transform(self.tgt2gosales["pandas"])
        for tgt, datasets in self.tgt2gosales.items():
            rasl_trained = rasl_trainable.fit(datasets)
            self._check_last_trained(sk_trained, rasl_trained, tgt)
            rasl_transformed = rasl_trained.transform(datasets)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(rasl_transformed), "index")
            rasl_transformed = _ensure_pandas(rasl_transformed)
            self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
            for row_idx in range(sk_transformed.shape[0]):
                for col_idx in range(sk_transformed.shape[1]):
                    self.assertEqual(
                        sk_transformed[row_idx, col_idx],
                        rasl_transformed.iloc[row_idx, col_idx],
                        (row_idx, col_idx, tgt),
                    )

    def test_predict(self):
        (train_X_pd, train_y_pd), (test_X_pd, test_y_pd) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        to_pd = Convert(astype="pandas")
        lr = LogisticRegression()
        encoder_args = {"handle_unknown": "use_encoded_value", "unknown_value": -1}
        sk_trainable = prefix >> SkOrdinalEncoder(**encoder_args) >> lr
        sk_trained = sk_trainable.fit(train_X_pd, train_y_pd)
        sk_predicted = sk_trained.predict(test_X_pd)
        rasl_trainable = prefix >> RaslOrdinalEncoder(**encoder_args) >> to_pd >> lr
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X, train_y)
            rasl_predicted = rasl_trained.predict(test_X)
            self.assertEqual(sk_predicted.shape, rasl_predicted.shape, tgt)
            self.assertEqual(sk_predicted.tolist(), rasl_predicted.tolist(), tgt)


def _check_trained_one_hot_encoder(test, op1, op2, msg):
    test.assertEqual(list(op1.feature_names_in_), list(op2.feature_names_in_), msg)
    test.assertEqual(len(op1.categories_), len(op2.categories_), msg)
    for i in range(len(op1.categories_)):
        test.assertEqual(list(op1.categories_[i]), list(op2.categories_[i]), msg)


class TestOneHotEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2creditg = cast(
            Dict[str, Any],
            {
                tgt: lale.datasets.openml.fetch(
                    "credit-g",
                    "classification",
                    preprocess=False,
                    astype=tgt,
                )
                for tgt in targets
            },
        )

    def _check_last_trained(self, op1, op2, msg):
        _check_trained_one_hot_encoder(
            self, op1.get_last().impl, op2.get_last().impl, msg
        )

    def test_fit(self):
        (train_X_pd, _), (_, _) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        rasl_trainable = prefix >> RaslOneHotEncoder()
        sk_trainable = prefix >> SkOneHotEncoder()
        sk_trained = sk_trainable.fit(train_X_pd)
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X)
            self._check_last_trained(sk_trained, rasl_trained, tgt)

    def test_partial_fit(self):
        (train_X_pd, _), (_, _) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        for tgt in self.tgt2creditg.keys():
            rasl_pipe = prefix >> RaslOneHotEncoder()
            for lower, upper in [[0, 10], [10, 100], [100, train_X_pd.shape[0]]]:
                data_so_far = train_X_pd[0:upper]
                sk_pipe = prefix >> SkOrdinalEncoder()
                sk_pipe = sk_pipe.fit(data_so_far)
                data_delta = train_X_pd[lower:upper]
                if tgt == "pandas":
                    pass
                elif tgt == "spark":
                    data_delta = pandas2spark(data_delta)
                elif tgt == "spark-with-index":
                    data_delta = pandas2spark(data_delta, with_index=True)
                else:
                    assert False
                rasl_pipe = rasl_pipe.partial_fit(data_delta)
                self._check_last_trained(
                    sk_pipe,
                    rasl_pipe,
                    (tgt, lower, upper),
                )

    def test_transform(self):
        (train_X_pd, train_y_pd), (test_X_pd, test_y_pd) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        rasl_trainable = prefix >> RaslOneHotEncoder(sparse=False)
        sk_trainable = prefix >> SkOneHotEncoder(sparse=False)
        sk_trained = sk_trainable.fit(train_X_pd)
        sk_transformed = sk_trained.transform(test_X_pd)
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X)
            self._check_last_trained(sk_trained, rasl_trained, tgt)
            rasl_transformed = rasl_trained.transform(test_X)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(rasl_transformed), "index")
            rasl_transformed = _ensure_pandas(rasl_transformed)
            self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
            for row_idx in range(sk_transformed.shape[0]):
                for col_idx in range(sk_transformed.shape[1]):
                    self.assertEqual(
                        sk_transformed[row_idx, col_idx],
                        rasl_transformed.iloc[row_idx, col_idx],
                        (row_idx, col_idx, tgt),
                    )

    def test_predict(self):
        (train_X_pd, train_y_pd), (test_X_pd, test_y_pd) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        to_pd = Convert(astype="pandas")
        lr = LogisticRegression()
        sk_trainable = prefix >> SkOneHotEncoder(sparse=False) >> lr
        sk_trained = sk_trainable.fit(train_X_pd, train_y_pd)
        sk_predicted = sk_trained.predict(test_X_pd)
        rasl_trainable = prefix >> RaslOneHotEncoder(sparse=False) >> to_pd >> lr
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X, train_y)
            rasl_predicted = rasl_trained.predict(test_X)
            self.assertEqual(sk_predicted.shape, rasl_predicted.shape, tgt)
            self.assertEqual(sk_predicted.tolist(), rasl_predicted.tolist(), tgt)


def _check_trained_hashing_encoder(test, op1, op2, msg):
    test.assertEqual(list(op1.feature_names), list(op2.feature_names), msg)


class TestHashingEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2creditg = cast(
            Dict[str, Any],
            {
                tgt: lale.datasets.openml.fetch(
                    "credit-g",
                    "classification",
                    preprocess=False,
                    astype=tgt,
                )
                for tgt in targets
            },
        )

    def _check_last_trained(self, op1, op2, msg):
        _check_trained_hashing_encoder(
            self, op1.get_last().impl, op2.get_last().impl, msg
        )

    def test_fit(self):
        (train_X_pd, _), (_, _) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        rasl_trainable = prefix >> RaslHashingEncoder()
        sk_trainable = prefix >> SkHashingEncoder()
        sk_trained = sk_trainable.fit(train_X_pd)
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X)
            self._check_last_trained(sk_trained, rasl_trained, tgt)

    def test_transform(self):
        (train_X_pd, train_y_pd), (test_X_pd, test_y_pd) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        rasl_trainable = prefix >> RaslHashingEncoder()
        sk_trainable = prefix >> SkHashingEncoder()
        sk_trained = sk_trainable.fit(train_X_pd)
        sk_transformed = sk_trained.transform(test_X_pd)
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X)
            self._check_last_trained(sk_trained, rasl_trained, tgt)
            rasl_transformed = rasl_trained.transform(test_X)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(rasl_transformed), "index")
            rasl_transformed = _ensure_pandas(rasl_transformed)
            self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
            for row_idx in range(sk_transformed.shape[0]):
                for col_idx in range(sk_transformed.shape[1]):
                    self.assertEqual(
                        sk_transformed.iloc[row_idx, col_idx],
                        rasl_transformed.iloc[row_idx, col_idx],
                        (row_idx, col_idx, tgt),
                    )

    def test_predict(self):
        (train_X_pd, train_y_pd), (test_X_pd, test_y_pd) = self.tgt2creditg["pandas"]
        cat_columns = categorical()(train_X_pd)
        prefix = Map(columns={c: it[c] for c in cat_columns})
        to_pd = Convert(astype="pandas")
        lr = LogisticRegression()
        sk_trainable = prefix >> SkHashingEncoder() >> lr
        sk_trained = sk_trainable.fit(train_X_pd, train_y_pd)
        sk_predicted = sk_trained.predict(test_X_pd)
        rasl_trainable = prefix >> RaslHashingEncoder() >> to_pd >> lr
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X, train_y)
            rasl_predicted = rasl_trained.predict(test_X)
            self.assertEqual(sk_predicted.shape, rasl_predicted.shape, tgt)
            self.assertEqual(sk_predicted.tolist(), rasl_predicted.tolist(), tgt)


class TestSimpleImputer(unittest.TestCase):
    def setUp(self):
        targets = ["pandas", "spark", "spark-with-index"]
        self.tgt2adult = {
            tgt: lale.datasets.openml.fetch(
                "adult",
                "classification",
                preprocess=False,
                astype=tgt,
            )
            for tgt in targets
        }

    def _fill_missing_value(self, col_name, value, missing_value):
        for tgt, datasets in self.tgt2adult.items():
            (train_X, train_y), (test_X, test_y) = datasets
            if tgt == "pandas":
                train_X.loc[
                    train_X[col_name] == value, col_name
                ] = missing_value  # type:ignore
                test_X.loc[
                    test_X[col_name] == value, col_name
                ] = missing_value  # type:ignore
            elif tgt.startswith("spark"):
                from pyspark.sql.functions import col, when

                train_X_new = train_X.withColumn(
                    col_name,
                    when(col(col_name) == value, missing_value).otherwise(
                        col(col_name)
                    ),
                )
                test_X_new = test_X.withColumn(
                    col_name,
                    when(col(col_name) == value, missing_value).otherwise(
                        col(col_name)
                    ),
                )
                train_X = forward_metadata(train_X, train_X_new)
                test_X = forward_metadata(test_X, test_X_new)
            else:
                assert False
            self.tgt2adult[tgt] = (train_X, train_y), (test_X, test_y)

    def test_fit_transform_numeric_nan_missing(self):
        self._fill_missing_value("age", 36.0, np.nan)
        num_columns = ["age", "fnlwgt", "education-num"]
        prefix = Map(columns={c: it[c] for c in num_columns})

        hyperparams = [
            {"strategy": "mean"},
            {"strategy": "median"},
            {"strategy": "most_frequent"},
            {"strategy": "constant", "fill_value": 99},
        ]
        for hyperparam in hyperparams:
            rasl_trainable = prefix >> RaslSimpleImputer(**hyperparam)
            sk_trainable = prefix >> SkSimpleImputer(**hyperparam)
            sk_trained = sk_trainable.fit(self.tgt2adult["pandas"][0][0])
            sk_transformed = sk_trained.transform(self.tgt2adult["pandas"][1][0])
            sk_statistics_ = sk_trained.steps[-1][1].impl.statistics_
            for tgt, dataset in self.tgt2adult.items():
                (train_X, _), (test_X, _) = dataset
                rasl_trained = rasl_trainable.fit(train_X)
                # test the fit succeeded.
                rasl_statistics_ = rasl_trained.steps[-1][1].impl.statistics_

                self.assertEqual(
                    len(sk_statistics_), len(rasl_statistics_), (hyperparam, tgt)
                )
                for i in range(sk_statistics_.shape[0]):
                    self.assertAlmostEqual(
                        sk_statistics_[i], rasl_statistics_[i], msg=(i, hyperparam, tgt)
                    )

                rasl_transformed = rasl_trained.transform(test_X)
                if tgt == "spark-with-index":
                    self.assertEqual(get_index_name(rasl_transformed), "index")
                rasl_transformed = _ensure_pandas(rasl_transformed)
                self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
                for row_idx in range(sk_transformed.shape[0]):
                    for col_idx in range(sk_transformed.shape[1]):
                        self.assertAlmostEqual(
                            sk_transformed[row_idx, col_idx],
                            rasl_transformed.iloc[row_idx, col_idx],
                            msg=(row_idx, col_idx, tgt),
                        )

    def test_fit_transform_numeric_nonan_missing(self):
        self._fill_missing_value("age", 36.0, -1)
        num_columns = ["age", "fnlwgt", "education-num"]
        prefix = Map(columns={c: it[c] for c in num_columns})

        hyperparams = [
            {"strategy": "mean"},
            {"strategy": "median"},
            {"strategy": "most_frequent"},
            {"strategy": "constant", "fill_value": 99},
        ]
        for hyperparam in hyperparams:
            rasl_trainable = prefix >> RaslSimpleImputer(
                missing_values=-1, **hyperparam
            )
            sk_trainable = prefix >> SkSimpleImputer(missing_values=-1, **hyperparam)
            sk_trained = sk_trainable.fit(self.tgt2adult["pandas"][0][0])
            sk_transformed = sk_trained.transform(self.tgt2adult["pandas"][1][0])
            sk_statistics_ = sk_trained.get_last().impl.statistics_
            for tgt, dataset in self.tgt2adult.items():
                (train_X, _), (test_X, _) = dataset
                rasl_trained = rasl_trainable.fit(train_X)
                # test the fit succeeded.
                rasl_statistics_ = rasl_trained.get_last().impl.statistics_  # type: ignore
                self.assertEqual(len(sk_statistics_), len(rasl_statistics_), tgt)
                self.assertEqual(list(sk_statistics_), list(rasl_statistics_), tgt)

                rasl_transformed = rasl_trained.transform(test_X)
                if tgt == "spark-with-index":
                    self.assertEqual(get_index_name(rasl_transformed), "index")
                rasl_transformed = _ensure_pandas(rasl_transformed)
                self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
                for row_idx in range(sk_transformed.shape[0]):
                    for col_idx in range(sk_transformed.shape[1]):
                        self.assertEqual(
                            sk_transformed[row_idx, col_idx],
                            rasl_transformed.iloc[row_idx, col_idx],
                            (row_idx, col_idx, tgt),
                        )

    def test_predict(self):
        self._fill_missing_value("age", 36.0, np.nan)
        (train_X_pd, train_y_pd), (test_X_pd, test_y_pd) = self.tgt2adult["pandas"]
        num_columns = ["age", "fnlwgt", "education-num"]
        prefix = Map(columns={c: it[c] for c in num_columns})
        to_pd = Convert(astype="pandas")
        lr = LogisticRegression()
        imputer_args = {"strategy": "mean"}
        sk_trainable = prefix >> SkSimpleImputer(**imputer_args) >> lr
        sk_trained = sk_trainable.fit(train_X_pd, train_y_pd)
        sk_predicted = sk_trained.predict(test_X_pd)
        rasl_trainable = prefix >> RaslSimpleImputer(**imputer_args) >> to_pd >> lr
        for tgt, dataset in self.tgt2adult.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X, train_y)
            rasl_predicted = rasl_trained.predict(test_X)
            self.assertEqual(sk_predicted.shape, rasl_predicted.shape, tgt)
            self.assertEqual(sk_predicted.tolist(), rasl_predicted.tolist(), tgt)

    def test_invalid_datatype_strategy(self):
        sk_trainable = SkSimpleImputer()
        with self.assertRaises(ValueError):
            sk_trainable.fit(self.tgt2adult["pandas"][0][0])
        rasl_trainable = RaslSimpleImputer()
        for tgt, dataset in self.tgt2adult.items():
            (train_X, _), (_, _) = dataset
            if tgt.startswith("spark"):
                # Skip test because of timeout!
                continue
            with self.assertRaises(ValueError):
                _ = rasl_trainable.fit(train_X)

    def test_default_numeric_fill_value(self):
        self._fill_missing_value("age", 36.0, np.nan)
        num_columns = ["age", "fnlwgt", "education-num"]
        prefix = Map(columns={c: it[c] for c in num_columns})

        hyperparams = [{"strategy": "constant"}]
        for hyperparam in hyperparams:
            rasl_trainable = prefix >> RaslSimpleImputer(**hyperparam)
            sk_trainable = prefix >> SkSimpleImputer(**hyperparam)
            sk_trained = sk_trainable.fit(self.tgt2adult["pandas"][0][0])
            sk_transformed = sk_trained.transform(self.tgt2adult["pandas"][1][0])
            sk_statistics_ = sk_trained.steps[-1][1].impl.statistics_
            for tgt, dataset in self.tgt2adult.items():
                (train_X, _), (test_X, _) = dataset
                rasl_trained = rasl_trainable.fit(train_X)
                # test the fit succeeded.
                rasl_statistics_ = rasl_trained.steps[-1][1].impl.statistics_
                self.assertEqual(len(sk_statistics_), len(rasl_statistics_), tgt)
                self.assertEqual(list(sk_statistics_), list(rasl_statistics_), tgt)

                rasl_transformed = rasl_trained.transform(test_X)
                rasl_transformed = _ensure_pandas(rasl_transformed)
                self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
                for row_idx in range(sk_transformed.shape[0]):
                    for col_idx in range(sk_transformed.shape[1]):
                        self.assertEqual(
                            sk_transformed[row_idx, col_idx],
                            rasl_transformed.iloc[row_idx, col_idx],
                            (row_idx, col_idx, tgt),
                        )

    def test_default_string_fill_value(self):
        self._fill_missing_value("education", "Prof-school", np.nan)

        str_columns = ["workclass", "education", "capital-gain"]
        prefix = Map(columns={c: it[c] for c in str_columns})

        hyperparams = [{"strategy": "constant"}]
        for hyperparam in hyperparams:
            rasl_trainable = prefix >> RaslSimpleImputer(**hyperparam)
            sk_trainable = prefix >> SkSimpleImputer(**hyperparam)
            sk_trained = sk_trainable.fit(self.tgt2adult["pandas"][0][0])
            sk_statistics_ = sk_trained.steps[-1][1].impl.statistics_
            for tgt, dataset in self.tgt2adult.items():
                (train_X, _), (test_X, _) = dataset
                rasl_trained = rasl_trainable.fit(train_X)
                # test the fit succeeded.
                rasl_statistics_ = rasl_trained.steps[-1][1].impl.statistics_
                self.assertEqual(len(sk_statistics_), len(rasl_statistics_), tgt)
                self.assertEqual(list(sk_statistics_), list(rasl_statistics_), tgt)

                rasl_transformed = rasl_trained.transform(test_X)
                rasl_transformed = _ensure_pandas(rasl_transformed)
                # Note that for this test case, the output of sklearn transform does not
                # match rasl transform. There is at least one row which has a None
                # value and pandas replace treats it as nan and replaces it.
                # Sklearn which uses numpy does not replace a None.
                # So we just test that `missing_value` is the default assigned.
                self.assertEqual(rasl_transformed.iloc[1, 1], "missing_value")

    def test_multiple_modes_numeric(self):
        # Sklearn SimpleImputer says: for strategy `most_frequent`,
        # if there is more than one such value, only the smallest is returned.
        data = [[1, 10], [2, 14], [3, 15], [4, 15], [5, 14], [6, np.nan]]
        df = pd.DataFrame(data, columns=["Id", "Age"])
        hyperparam = {"strategy": "most_frequent"}
        sk_trainable = SkSimpleImputer(**hyperparam)
        rasl_trainable = RaslSimpleImputer(**hyperparam)
        sk_trained = sk_trainable.fit(df)
        rasl_trained = rasl_trainable.fit(df)
        self.assertEqual(
            len(sk_trained.statistics_), len(rasl_trained.impl.statistics_), "pandas"
        )
        self.assertEqual([6, 15], list(rasl_trained.impl.statistics_), "pandas")
        from pyspark.sql import SparkSession

        spark = (
            SparkSession.builder.master("local[1]")
            .appName("test_relational_sklearn")
            .getOrCreate()
        )
        spark_df = spark.createDataFrame(df)  # type:ignore

        rasl_trained = rasl_trainable.fit(spark_df)
        self.assertEqual(
            len(sk_trained.statistics_), len(rasl_trained.impl.statistics_), "spark"
        )
        self.assertIn(rasl_trained.impl.statistics_[1], [14, 15])

    def test_multiple_modes_string(self):
        # Sklearn SimpleImputer says: for strategy `most_frequent`,
        # if there is more than one such value, only the smallest is returned.
        data = [
            ["a", "t"],
            ["b", "f"],
            ["b", "m"],
            ["c", "f"],
            ["c", "m"],
            ["f", "missing"],
        ]
        df = pd.DataFrame(data, columns=["Id", "Gender"])
        hyperparam = {"strategy": "most_frequent", "missing_values": "missing"}
        sk_trainable = SkSimpleImputer(**hyperparam)
        rasl_trainable = RaslSimpleImputer(**hyperparam)
        sk_trained = sk_trainable.fit(df)
        rasl_trained = rasl_trainable.fit(df)
        self.assertEqual(
            len(sk_trained.statistics_), len(rasl_trained.impl.statistics_), "pandas"
        )
        self.assertEqual(
            list(["c", "m"]), list(rasl_trained.impl.statistics_), "pandas"
        )

        from pyspark.sql import SparkSession

        spark = (
            SparkSession.builder.master("local[1]")
            .appName("test_relational_sklearn")
            .getOrCreate()
        )
        spark_df = spark.createDataFrame(df)  # type:ignore

        rasl_trained = rasl_trainable.fit(spark_df)
        self.assertEqual(
            len(sk_trained.statistics_), len(rasl_trained.impl.statistics_), "spark"
        )
        self.assertIn(rasl_trained.impl.statistics_[1], ["f", "m"])

    def test_valid_partial_fit(self):
        self._fill_missing_value("age", 36.0, -1)
        num_columns = ["age", "fnlwgt", "education-num"]
        prefix = Map(columns={c: it[c] for c in num_columns})

        hyperparams = [
            {"strategy": "mean"},
            {"strategy": "constant", "fill_value": 99},
        ]
        for hyperparam in hyperparams:
            rasl_trainable = prefix >> RaslSimpleImputer(
                missing_values=-1, **hyperparam
            )
            sk_trainable = prefix >> SkSimpleImputer(missing_values=-1, **hyperparam)
            sk_trained = sk_trainable.fit(self.tgt2adult["pandas"][0][0])
            sk_transformed = sk_trained.transform(self.tgt2adult["pandas"][1][0])
            sk_statistics_ = sk_trained.get_last().impl.statistics_
            (train_X, _), (test_X, _) = self.tgt2adult["pandas"]
            data1_pandas = train_X.iloc[:10]
            data2_pandas = train_X.iloc[10:100]
            data3_pandas = train_X.iloc[100:]
            test_X_pandas = test_X

            for tgt in self.tgt2adult.keys():
                if tgt == "pandas":
                    data1 = data1_pandas
                    data2 = data2_pandas
                    data3 = data3_pandas
                    test_X = test_X_pandas
                elif tgt.startswith("spark"):
                    with_index = tgt == "spark-with-index"
                    data1 = pandas2spark(data1_pandas, with_index)  # type:ignore
                    data2 = pandas2spark(data2_pandas, with_index)  # type:ignore
                    data3 = pandas2spark(data3_pandas, with_index)  # type:ignore
                    test_X = pandas2spark(test_X_pandas, with_index)  # type:ignore
                else:
                    assert False
                rasl_trainable = prefix >> RaslSimpleImputer(
                    missing_values=-1, **hyperparam
                )
                rasl_trained = rasl_trainable.partial_fit(data1)
                rasl_trained = rasl_trained.partial_fit(data2)
                rasl_trained = rasl_trained.partial_fit(data3)
                # test the fit succeeded.
                rasl_statistics_ = rasl_trained.get_last().impl.statistics_  # type: ignore

                self.assertEqual(len(sk_statistics_), len(rasl_statistics_), tgt)
                for i in range(len(sk_statistics_)):
                    self.assertEqual(sk_statistics_[i], rasl_statistics_[i])

                rasl_transformed = rasl_trained.transform(test_X)
                rasl_transformed = _ensure_pandas(rasl_transformed)
                self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
                for row_idx in range(sk_transformed.shape[0]):
                    for col_idx in range(sk_transformed.shape[1]):
                        self.assertEqual(
                            sk_transformed[row_idx, col_idx],
                            rasl_transformed.iloc[row_idx, col_idx],
                            (row_idx, col_idx, tgt),
                        )

    def test_invalid_partial_fit(self):
        num_columns = ["age", "fnlwgt", "education-num"]
        prefix = Map(columns={c: it[c] for c in num_columns})

        hyperparams = [
            {"strategy": "median"},
            {"strategy": "most_frequent"},
        ]
        for hyperparam in hyperparams:
            rasl_trainable = prefix >> RaslSimpleImputer(
                missing_values=-1, **hyperparam
            )
            (train_X, _), (_, _) = self.tgt2adult["pandas"]
            with self.assertRaises(ValueError):
                _ = rasl_trainable.partial_fit(train_X)


def _check_trained_standard_scaler(test, op1, op2, msg):
    test.assertEqual(list(op1.feature_names_in_), list(op2.feature_names_in_), msg)
    test.assertEqual(op1.n_features_in_, op2.n_features_in_, msg)
    test.assertEqual(op1.n_samples_seen_, op2.n_samples_seen_, msg)
    if op1.mean_ is None:
        test.assertIsNone(op2.mean_, msg)
    else:
        test.assertIsNotNone(op2.mean_, msg)
        test.assertEqual(len(op1.mean_), len(op2.mean_), msg)
        for i in range(len(op1.mean_)):
            test.assertAlmostEqual(op1.mean_[i], op2.mean_[i], msg=msg)
    if op1.var_ is None:
        test.assertIsNone(op2.var_, msg)
    else:
        test.assertIsNotNone(op2.var_, msg)
        test.assertEqual(len(op1.var_), len(op2.var_), msg)
        for i in range(len(op1.var_)):
            test.assertAlmostEqual(op1.var_[i], op2.var_[i], msg=msg)
    if op1.scale_ is None:
        test.assertIsNone(op2.scale_, msg)
    else:
        test.assertIsNotNone(op2.scale_, msg)
        test.assertEqual(len(op1.scale_), len(op2.scale_), msg)
        for i in range(len(op1.scale_)):
            test.assertAlmostEqual(op1.scale_[i], op2.scale_[i], msg=msg)


class TestStandardScaler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2creditg = cast(
            Dict[str, Any],
            {
                tgt: lale.datasets.openml.fetch(
                    "credit-g",
                    "classification",
                    preprocess=True,
                    astype=tgt,
                )
                for tgt in targets
            },
        )

    def test_fit(self):
        (train_X_pd, _), (_, _) = self.tgt2creditg["pandas"]
        sk_trainable = SkStandardScaler()
        sk_trained = sk_trainable.fit(train_X_pd)
        rasl_trainable = RaslStandardScaler()
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, _), (_, _) = dataset
            rasl_trained = rasl_trainable.fit(train_X)
            _check_trained_standard_scaler(self, sk_trained, rasl_trained.impl, tgt)

    def test_partial_fit(self):
        (train_X_pd, _), (_, _) = self.tgt2creditg["pandas"]
        for tgt in self.tgt2creditg.keys():
            rasl_op = RaslStandardScaler()
            for lower, upper in [[0, 10], [10, 100], [100, train_X_pd.shape[0]]]:
                data_so_far = train_X_pd[0:upper]
                sk_op = SkStandardScaler()
                sk_op = sk_op.fit(data_so_far)
                data_delta = train_X_pd[lower:upper]
                if tgt == "pandas":
                    pass
                elif tgt == "spark":
                    data_delta = pandas2spark(data_delta)
                elif tgt == "spark-with-index":
                    data_delta = pandas2spark(data_delta, with_index=True)
                else:
                    assert False
                rasl_op = rasl_op.partial_fit(data_delta)
                _check_trained_standard_scaler(
                    self, sk_op, rasl_op.impl, (tgt, lower, upper)
                )

    def test_transform(self):
        (train_X_pd, _), (test_X_pd, _) = self.tgt2creditg["pandas"]
        sk_trainable = SkStandardScaler()
        sk_trained = sk_trainable.fit(train_X_pd)
        sk_transformed = sk_trained.transform(test_X_pd)
        rasl_trainable = RaslStandardScaler()
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, _), (test_X, _) = dataset
            rasl_trained = rasl_trainable.fit(train_X)
            _check_trained_standard_scaler(self, sk_trained, rasl_trained.impl, tgt)
            rasl_transformed = rasl_trained.transform(test_X)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(rasl_transformed), "index")
            rasl_transformed = _ensure_pandas(rasl_transformed)
            self.assertEqual(sk_transformed.shape, rasl_transformed.shape, tgt)
            for row_idx in range(sk_transformed.shape[0]):
                for col_idx in range(sk_transformed.shape[1]):
                    self.assertAlmostEqual(
                        sk_transformed[row_idx, col_idx],
                        rasl_transformed.iloc[row_idx, col_idx],
                        msg=(row_idx, col_idx, tgt),
                    )

    def test_predict(self):
        (train_X_pd, train_y_pd), (test_X_pd, test_y_pd) = self.tgt2creditg["pandas"]
        to_pd = Convert(astype="pandas")
        lr = LogisticRegression()
        sk_trainable = SkStandardScaler() >> lr
        sk_trained = sk_trainable.fit(train_X_pd, train_y_pd)
        sk_predicted = sk_trained.predict(test_X_pd)
        rasl_trainable = RaslStandardScaler() >> to_pd >> lr
        for tgt, dataset in self.tgt2creditg.items():
            (train_X, train_y), (test_X, test_y) = dataset
            rasl_trained = rasl_trainable.fit(train_X, train_y)
            rasl_predicted = rasl_trained.predict(test_X)
            self.assertEqual(sk_predicted.shape, rasl_predicted.shape, tgt)
            self.assertEqual(sk_predicted.tolist(), rasl_predicted.tolist(), tgt)


class _BatchTestingKFold:
    def __init__(self, n_batches, n_splits):
        self.n_batches = n_batches
        self.n_splits = n_splits

    def get_n_splits(self, X=None, y=None, groups=None):
        return self.n_splits

    def split(self, X, y=None, groups=None):
        # re-arrange batched data [[d0,e0,f0], [d1,e1,f1], [d2,e2,f2]]
        # into non-batched data [d0+d1+d2, e0+e1+e2, f0+f1+f2]
        result = [([], []) for _ in range(self.n_splits)]
        cv = KFold(self.n_splits)
        batches = mockup_data_loader(X, y, self.n_batches, "pandas")
        for idx, (bX, by) in enumerate(batches):
            for fold, (_, test) in enumerate(cv.split(bX, by)):
                remapped = bX.index[test]
                for f in range(self.n_splits):
                    if f != fold:
                        assert set(result[f][0]).isdisjoint(set(remapped))
                        result[f][0].extend(remapped)
                assert set(result[fold][1]).isdisjoint(set(remapped))
                result[fold][1].extend(remapped)
        return result


class _BatchTestingCallback:
    def __init__(self):
        self.n_calls = 0

    def __call__(self, score, n_batches_scanned, end_of_scanned_batches):
        self.n_calls += 1
        assert self.n_calls == n_batches_scanned, (self.n_calls, n_batches_scanned)
        assert not end_of_scanned_batches


class TestTaskGraphs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        X, y, fairness_info = lale.lib.aif360.fetch_creditg_df(preprocess=False)
        X = Project(columns=categorical()).fit(X).transform(X)
        fairness_info = {  # remove numeric protected attribute age
            "favorable_labels": fairness_info["favorable_labels"],
            "protected_attributes": fairness_info["protected_attributes"][:1],
        }
        cls.creditg = X, y, fairness_info

    @classmethod
    def _make_sk_trainable(cls, final_est):
        if final_est == "sgd":
            est = SGDClassifier(random_state=97)
        elif final_est == "rfc":
            est = RandomForestClassifier(random_state=97)
        else:
            assert False, final_est
        return sk_make_pipeline(
            SkOrdinalEncoder(handle_unknown="use_encoded_value", unknown_value=-1),
            SkMinMaxScaler(),
            est,
        )

    @classmethod
    def _make_rasl_trainable(cls, final_est):
        if final_est == "sgd":
            est = SGDClassifier(random_state=97)
        elif final_est == "rfc":
            est = RandomForestClassifier(random_state=97)
        else:
            assert False, final_est
        return (
            RaslOrdinalEncoder(handle_unknown="use_encoded_value", unknown_value=-1)
            >> RaslMinMaxScaler()
            >> est
        )

    def test_fit_no_batching(self):
        train_X, train_y, _ = self.creditg
        sk_trainable = self._make_sk_trainable("sgd")
        sk_trained = sk_trainable.fit(train_X, train_y)
        rasl_trainable = self._make_rasl_trainable("sgd")
        rasl_trained = rasl_trainable.fit(train_X, train_y)
        _check_trained_ordinal_encoder(
            self, sk_trained.steps[0][1], rasl_trained.steps[0][1].impl, "pandas"
        )
        _check_trained_min_max_scaler(
            self, sk_trained.steps[1][1], rasl_trained.steps[1][1].impl, "pandas"
        )

    def test_fit_batching(self):
        train_X, train_y, _ = self.creditg
        train_data_space = train_X.memory_usage().sum() + train_y.memory_usage()
        sk_trainable = self._make_sk_trainable("sgd")
        sk_trained = sk_trainable.fit(train_X, train_y)
        unique_class_labels = list(train_y.unique())
        for n_batches in [1, 3]:
            for prio in [PrioStep(), PrioBatch()]:
                batches = mockup_data_loader(train_X, train_y, n_batches, "pandas")
                rasl_trainable = self._make_rasl_trainable("sgd")
                rasl_trained = fit_with_batches(
                    pipeline=rasl_trainable,
                    batches=batches,
                    scoring=None,
                    unique_class_labels=unique_class_labels,
                    max_resident=3 * math.ceil(train_data_space / n_batches),
                    prio=prio,
                    partial_transform=False,
                    verbose=0,
                    progress_callback=None,
                )
                _check_trained_ordinal_encoder(
                    self,
                    sk_trained.steps[0][1],
                    rasl_trained.steps[0][1].impl,
                    (n_batches, type(prio)),
                )
                _check_trained_min_max_scaler(
                    self,
                    sk_trained.steps[1][1],
                    rasl_trained.steps[1][1].impl,
                    (n_batches, type(prio)),
                )

    def test_partial_transform(self):
        train_X, train_y, _ = self.creditg
        unique_class_labels = list(train_y.unique())
        for n_batches in [1, 3]:
            batches = mockup_data_loader(train_X, train_y, n_batches, "pandas")
            rasl_trainable = self._make_rasl_trainable("sgd")
            progress_callback = _BatchTestingCallback()
            _ = fit_with_batches(
                pipeline=rasl_trainable,
                batches=batches,
                scoring=rasl_get_scorer("accuracy"),
                unique_class_labels=unique_class_labels,
                max_resident=None,
                prio=PrioBatch(),
                partial_transform=True,
                verbose=0,
                progress_callback=progress_callback,
            )
            self.assertEqual(progress_callback.n_calls, n_batches)

    @unittest.skip("TODO: make sure SGD partial_fit i happens before scan i+1")
    def test_prefix_freeze_trained(self):
        train_X, train_y, _ = self.creditg
        unique_class_labels = list(train_y.unique())
        n_batches = 3
        batches = mockup_data_loader(train_X, train_y, n_batches, "pandas")
        first_batch = next(iter(batches))
        trainable1 = self._make_rasl_trainable("sgd")
        prefix1, suffix1 = trainable1.remove_last(), trainable1.get_last()
        prefix2 = fit_with_batches(
            pipeline=prefix1,
            batches=[first_batch],
            scoring=None,
            unique_class_labels=unique_class_labels,
            max_resident=None,
            prio=PrioBatch(),
            partial_transform=False,
            verbose=0,
            progress_callback=None,
        )
        trainable2 = prefix2.freeze_trained() >> suffix1
        progress_callback = _BatchTestingCallback()
        _ = fit_with_batches(
            pipeline=trainable2,
            batches=batches,
            scoring=rasl_get_scorer("accuracy"),
            unique_class_labels=unique_class_labels,
            max_resident=None,
            prio=PrioBatch(),
            partial_transform=False,
            verbose=0,
            progress_callback=progress_callback,
        )
        self.assertEqual(progress_callback.n_calls, n_batches - 1)

    def test_cross_val_score_accuracy(self):
        X, y, _ = self.creditg
        n_splits = 3
        for n_batches in [1, 3]:
            with self.assertWarnsRegex(DeprecationWarning, "trainable operator"):
                sk_scores = sk_cross_val_score(
                    estimator=self._make_sk_trainable("rfc"),
                    X=X,
                    y=y,
                    scoring=make_scorer(sk_accuracy_score),
                    cv=_BatchTestingKFold(n_batches, n_splits),
                )
            rasl_scores = rasl_cross_val_score(
                pipeline=self._make_rasl_trainable("rfc"),
                batches=mockup_data_loader(X, y, n_batches, "pandas"),
                scoring=rasl_get_scorer("accuracy"),
                cv=KFold(n_splits),
                unique_class_labels=list(y.unique()),
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                verbose=0,
            )
            for sk_s, rasl_s in zip(sk_scores, rasl_scores):
                self.assertAlmostEqual(sk_s, rasl_s, msg=n_batches)

    def test_cross_val_score_disparate_impact(self):
        X, y, fairness_info = self.creditg
        disparate_impact_scorer = lale.lib.aif360.disparate_impact(**fairness_info)
        n_splits = 3
        for n_batches in [1, 3]:
            with self.assertWarnsRegex(DeprecationWarning, "trainable operator"):
                sk_scores = sk_cross_val_score(
                    estimator=self._make_sk_trainable("rfc"),
                    X=X,
                    y=y,
                    scoring=disparate_impact_scorer,
                    cv=_BatchTestingKFold(n_batches, n_splits),
                )
            rasl_scores = rasl_cross_val_score(
                pipeline=self._make_rasl_trainable("rfc"),
                batches=mockup_data_loader(X, y, n_batches, "pandas"),
                scoring=disparate_impact_scorer,
                cv=KFold(n_splits),
                unique_class_labels=list(y.unique()),
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                verbose=0,
            )
            for sk_s, rasl_s in zip(sk_scores, rasl_scores):
                self.assertAlmostEqual(sk_s, rasl_s, msg=n_batches)

    def test_cross_validate(self):
        X, y, _ = self.creditg
        n_splits = 3
        for n_batches in [1, 3]:
            with self.assertWarnsRegex(DeprecationWarning, "trainable operator"):
                sk_scr = sk_cross_validate(
                    estimator=self._make_sk_trainable("rfc"),
                    X=X,
                    y=y,
                    scoring=make_scorer(sk_accuracy_score),
                    cv=_BatchTestingKFold(n_batches, n_splits),
                    return_estimator=True,
                )
            rasl_scr = rasl_cross_validate(
                pipeline=self._make_rasl_trainable("rfc"),
                batches=mockup_data_loader(X, y, n_batches, "pandas"),
                scoring=rasl_get_scorer("accuracy"),
                cv=KFold(n_splits),
                unique_class_labels=list(y.unique()),
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                return_estimator=True,
                verbose=0,
            )
            for sk_s, rasl_s in zip(sk_scr["test_score"], rasl_scr["test_score"]):
                self.assertAlmostEqual(cast(float, sk_s), cast(float, rasl_s))
            for sk_e, rasl_e in zip(sk_scr["estimator"], rasl_scr["estimator"]):
                _check_trained_ordinal_encoder(
                    self,
                    sk_e.steps[0][1],
                    cast(TrainedPipeline, rasl_e).steps[0][1].impl,
                    n_batches,
                )
                _check_trained_min_max_scaler(
                    self,
                    sk_e.steps[1][1],
                    cast(TrainedPipeline, rasl_e).steps[1][1].impl,
                    n_batches,
                )


class TestTaskGraphsWithConcat(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
            "credit-g", "classification", preprocess=False
        )
        cls.cat_columns = [
            "checking_status",
            "credit_history",
            "purpose",
            "savings_status",
            "employment",
            "personal_status",
            "other_parties",
            "property_magnitude",
            "other_payment_plans",
            "housing",
            "job",
            "own_telephone",
            "foreign_worker",
        ]
        cls.num_columns = [
            "duration",
            "credit_amount",
            "installment_commitment",
            "residence_since",
            "age",
            "existing_credits",
            "num_dependents",
        ]
        cls.creditg = (train_X, train_y), (test_X, test_y)

    @classmethod
    def _make_sk_trainable(cls, final_est):
        from sklearn.compose import ColumnTransformer
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.linear_model import SGDClassifier

        if final_est == "sgd":
            est = SGDClassifier(random_state=97)
        elif final_est == "rfc":
            est = RandomForestClassifier(random_state=97)
        else:
            assert False, final_est
        return sk_make_pipeline(
            ColumnTransformer(
                [
                    (
                        "prep_cat",
                        SkOrdinalEncoder(
                            handle_unknown="use_encoded_value", unknown_value=-1
                        ),
                        cls.cat_columns,
                    ),
                    (
                        "prep_num",
                        sk_make_pipeline(
                            SkSimpleImputer(strategy="mean"),
                            SkMinMaxScaler(),
                            "passthrough",
                        ),
                        cls.num_columns,
                    ),
                ]
            ),
            est,
        )

    @classmethod
    def _make_rasl_trainable(cls, final_est):
        if final_est == "sgd":
            est = SGDClassifier(random_state=97)
        elif final_est == "rfc":
            est = RandomForestClassifier(random_state=97)
        else:
            assert False, final_est
        return (
            (
                (
                    Project(columns=cls.cat_columns)
                    >> RaslOrdinalEncoder(
                        handle_unknown="use_encoded_value", unknown_value=-1
                    )
                )
                & (
                    Project(columns=cls.num_columns)
                    >> RaslSimpleImputer(strategy="mean")
                    >> RaslMinMaxScaler()
                )
            )
            >> ConcatFeatures()
            >> est
        )

    def test_fit_no_batching(self):
        (train_X, train_y), _ = self.creditg
        sk_trainable = self._make_sk_trainable("sgd")
        sk_trained = sk_trainable.fit(train_X, train_y)
        rasl_trainable = self._make_rasl_trainable("sgd")
        rasl_trained = rasl_trainable.fit(train_X, train_y)
        _check_trained_ordinal_encoder(
            self,
            sk_trained.steps[0][1].transformers_[0][1],
            rasl_trained.steps[1][1].impl,
            "pandas",
        )
        _check_trained_min_max_scaler(
            self,
            sk_trained.steps[0][1].transformers_[1][1].steps[1][1],
            rasl_trained.steps[4][1].impl,
            "pandas",
        )

    def test_fit_batching(self):
        (train_X, train_y), _ = self.creditg
        train_data_space = train_X.memory_usage().sum() + train_y.memory_usage()
        sk_trainable = self._make_sk_trainable("sgd")
        sk_trained = sk_trainable.fit(train_X, train_y)
        unique_class_labels = list(train_y.unique())
        for n_batches in [1, 3]:
            for prio in [PrioStep(), PrioBatch()]:
                batches = create_data_loader(
                    train_X, train_y, math.ceil(len(train_y) / n_batches)
                )
                self.assertEqual(n_batches, len(batches))
                rasl_trainable = self._make_rasl_trainable("sgd")
                rasl_trained = fit_with_batches(
                    pipeline=rasl_trainable,
                    batches=batches,  # type: ignore
                    scoring=None,
                    unique_class_labels=unique_class_labels,
                    max_resident=3 * math.ceil(train_data_space / n_batches),
                    prio=prio,
                    partial_transform=False,
                    verbose=0,
                    progress_callback=None,
                )
                _check_trained_ordinal_encoder(
                    self,
                    sk_trained.steps[0][1].transformers_[0][1],
                    rasl_trained.steps[1][1].impl,
                    (n_batches, type(prio)),
                )
                _check_trained_min_max_scaler(
                    self,
                    sk_trained.steps[0][1].transformers_[1][1].steps[1][1],
                    rasl_trained.steps[4][1].impl,
                    (n_batches, type(prio)),
                )

    def test_cross_val_score(self):
        (X, y), _ = self.creditg
        sk_scores = sk_cross_val_score(
            estimator=self._make_sk_trainable("rfc"),
            X=X,
            y=y,
            scoring=make_scorer(sk_accuracy_score),
            cv=KFold(3),
        )
        for n_batches in [1, 3]:
            rasl_scores = rasl_cross_val_score(
                pipeline=self._make_rasl_trainable("rfc"),
                batches=mockup_data_loader(X, y, n_batches, "pandas"),
                scoring=rasl_get_scorer("accuracy"),
                cv=KFold(3),
                unique_class_labels=list(y.unique()),
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                verbose=0,
            )
            if n_batches == 1:
                for sk_s, rasl_s in zip(sk_scores, rasl_scores):
                    self.assertAlmostEqual(sk_s, rasl_s)


class TestTaskGraphsWithCategoricalConcat(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        (train_X, train_y), (test_X, test_y) = lale.datasets.openml.fetch(
            "credit-g", "classification", preprocess=False
        )
        cls.cat_columns = [
            "checking_status",
            "credit_history",
            "purpose",
            "savings_status",
            "employment",
            "personal_status",
            "other_parties",
            "property_magnitude",
            "other_payment_plans",
            "housing",
            "job",
            "own_telephone",
            "foreign_worker",
        ]
        cls.num_columns = [
            "duration",
            "credit_amount",
            "installment_commitment",
            "residence_since",
            "age",
            "existing_credits",
            "num_dependents",
        ]
        cls.creditg = (train_X, train_y), (test_X, test_y)

    @classmethod
    def _make_sk_trainable(cls, final_est):
        from sklearn.compose import ColumnTransformer
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.linear_model import SGDClassifier

        if final_est == "sgd":
            est = SGDClassifier(random_state=97)
        elif final_est == "rfc":
            est = RandomForestClassifier(random_state=97)
        else:
            assert False, final_est
        return sk_make_pipeline(
            ColumnTransformer(
                [
                    (
                        "prep_cat",
                        SkOrdinalEncoder(
                            handle_unknown="use_encoded_value", unknown_value=-1
                        ),
                        cls.cat_columns,
                    ),
                    (
                        "prep_num",
                        sk_make_pipeline(
                            SkSimpleImputer(strategy="mean"),
                            SkMinMaxScaler(),
                            "passthrough",
                        ),
                        cls.num_columns,
                    ),
                ]
            ),
            est,
        )

    @classmethod
    def _make_rasl_trainable(cls, final_est):
        if final_est == "sgd":
            est = SGDClassifier(random_state=97)
        elif final_est == "rfc":
            est = RandomForestClassifier(random_state=97)
        else:
            assert False, final_est
        return (
            (
                (
                    Project(columns=categorical(11), drop_columns={"type": "number"})
                    >> RaslOrdinalEncoder(
                        handle_unknown="use_encoded_value", unknown_value=-1
                    )
                )
                & (
                    Project(columns={"type": "number"})
                    >> RaslSimpleImputer(strategy="mean")
                    >> RaslMinMaxScaler()
                )
            )
            >> ConcatFeatures()
            >> est
        )

    def test_fit_no_batching(self):
        (train_X, train_y), _ = self.creditg
        sk_trainable = self._make_sk_trainable("sgd")
        sk_trained = sk_trainable.fit(train_X, train_y)
        rasl_trainable = self._make_rasl_trainable("sgd")
        rasl_trained = rasl_trainable.fit(train_X, train_y)
        _check_trained_ordinal_encoder(
            self,
            sk_trained.steps[0][1].transformers_[0][1],
            rasl_trained.steps[1][1].impl,
            "pandas",
        )
        _check_trained_min_max_scaler(
            self,
            sk_trained.steps[0][1].transformers_[1][1].steps[1][1],
            rasl_trained.steps[4][1].impl,
            "pandas",
        )

    def test_fit_batching(self):
        (train_X, train_y), _ = self.creditg
        train_data_space = train_X.memory_usage().sum() + train_y.memory_usage()
        sk_trainable = self._make_sk_trainable("sgd")
        sk_trained = sk_trainable.fit(train_X, train_y)
        unique_class_labels = list(train_y.unique())
        for n_batches in [1, 3]:
            for prio in [PrioStep(), PrioBatch()]:
                batches = create_data_loader(
                    train_X, train_y, math.ceil(len(train_y) / n_batches)
                )
                self.assertEqual(n_batches, len(batches))
                rasl_trainable = self._make_rasl_trainable("sgd")
                rasl_trained = fit_with_batches(
                    pipeline=rasl_trainable,
                    batches=batches,  # type: ignore
                    scoring=None,
                    unique_class_labels=unique_class_labels,
                    max_resident=3 * math.ceil(train_data_space / n_batches),
                    prio=prio,
                    partial_transform=False,
                    verbose=0,
                    progress_callback=None,
                )
                _check_trained_ordinal_encoder(
                    self,
                    sk_trained.steps[0][1].transformers_[0][1],
                    rasl_trained.steps[1][1].impl,
                    (n_batches, type(prio)),
                )
                _check_trained_min_max_scaler(
                    self,
                    sk_trained.steps[0][1].transformers_[1][1].steps[1][1],
                    rasl_trained.steps[4][1].impl,
                    (n_batches, type(prio)),
                )

    def test_cross_val_score(self):
        (X, y), _ = self.creditg
        sk_scores = sk_cross_val_score(
            estimator=self._make_sk_trainable("rfc"),
            X=X,
            y=y,
            scoring=make_scorer(sk_accuracy_score),
            cv=KFold(3),
        )
        for n_batches in [1, 3]:
            rasl_scores = rasl_cross_val_score(
                pipeline=self._make_rasl_trainable("rfc"),
                batches=mockup_data_loader(X, y, n_batches, "pandas"),
                scoring=rasl_get_scorer("accuracy"),
                cv=KFold(3),
                unique_class_labels=list(y.unique()),
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                verbose=0,
            )
            if n_batches == 1:
                for sk_s, rasl_s in zip(sk_scores, rasl_scores):
                    self.assertAlmostEqual(sk_s, rasl_s)


class TestTaskGraphsSpark(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        X, y, _ = lale.lib.aif360.fetch_creditg_df(preprocess=False)
        X = Project(columns=categorical()).fit(X).transform(X)
        cls.creditg = X, y

    @classmethod
    def _make_sk_trainable(cls):
        return sk_make_pipeline(
            SkOrdinalEncoder(handle_unknown="use_encoded_value", unknown_value=-1),
            SkMinMaxScaler(),
            Convert(astype="pandas"),
            RandomForestClassifier(random_state=97),
        )

    @classmethod
    def _make_rasl_trainable(cls):
        return (
            RaslOrdinalEncoder(handle_unknown="use_encoded_value", unknown_value=-1)
            >> RaslMinMaxScaler()
            >> Convert(astype="pandas")
            >> RandomForestClassifier(random_state=97)
        )

    def test_fit_with_batches(self):
        train_X, train_y = self.creditg
        sk_trainable = self._make_sk_trainable()
        sk_trained = sk_trainable.fit(train_X, train_y)
        unique_class_labels = list(train_y.unique())
        for tgt, n_batches in itertools.product(["pandas", "spark"], [1, 3]):
            rasl_trained = fit_with_batches(
                pipeline=self._make_rasl_trainable(),
                batches=mockup_data_loader(train_X, train_y, n_batches, tgt),
                scoring=None,
                unique_class_labels=unique_class_labels,
                max_resident=None,
                prio=PrioBatch(),
                partial_transform=False,
                verbose=0,
                progress_callback=None,
            )
            _check_trained_ordinal_encoder(
                self, sk_trained.steps[0][1], rasl_trained.steps[0][1].impl, tgt
            )
            _check_trained_min_max_scaler(
                self, sk_trained.steps[1][1], rasl_trained.steps[1][1].impl, tgt
            )

    def test_cross_val_score(self):
        X, y = self.creditg
        with self.assertWarnsRegex(DeprecationWarning, "trainable operator"):
            sk_scores = sk_cross_val_score(
                estimator=self._make_sk_trainable(),
                X=X,
                y=y,
                scoring=make_scorer(sk_accuracy_score),
                cv=KFold(3),
            )
        unique_class_labels = list(y.unique())
        for tgt, n_batches in itertools.product(["pandas", "spark"], [1, 3]):
            rasl_scores = rasl_cross_val_score(
                pipeline=self._make_rasl_trainable(),
                batches=mockup_data_loader(X, y, n_batches, tgt),
                scoring=rasl_get_scorer("accuracy"),
                cv=KFold(3),
                unique_class_labels=unique_class_labels,
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                verbose=0,
            )
            if n_batches == 1:
                for sk_s, rasl_s in zip(sk_scores, rasl_scores):
                    self.assertAlmostEqual(sk_s, rasl_s, msg=(tgt, n_batches))

    def test_cross_validate(self):
        X, y = self.creditg
        with self.assertWarnsRegex(DeprecationWarning, "trainable operator"):
            sk_scores = sk_cross_validate(
                estimator=self._make_sk_trainable(),
                X=X,
                y=y,
                scoring=make_scorer(sk_accuracy_score),
                cv=KFold(3),
                return_estimator=True,
            )
        unique_class_labels = list(y.unique())
        for tgt, n_batches in itertools.product(["pandas", "spark"], [1, 3]):
            rasl_scores = rasl_cross_validate(
                pipeline=self._make_rasl_trainable(),
                batches=mockup_data_loader(X, y, n_batches, tgt),
                scoring=rasl_get_scorer("accuracy"),
                cv=KFold(3),
                unique_class_labels=unique_class_labels,
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                return_estimator=True,
                verbose=0,
            )
            msg = tgt, n_batches
            if n_batches == 1:
                for sk_s, rasl_s in zip(
                    sk_scores["test_score"], rasl_scores["test_score"]
                ):
                    self.assertAlmostEqual(
                        cast(float, sk_s), cast(float, rasl_s), msg=msg
                    )
                for sk_e, rasl_e in zip(
                    sk_scores["estimator"], rasl_scores["estimator"]
                ):
                    rasl_steps = cast(TrainedPipeline, rasl_e).steps
                    _check_trained_ordinal_encoder(
                        self,
                        sk_e.steps[0][1],
                        rasl_steps[0][1].impl,
                        msg=msg,
                    )
                    _check_trained_min_max_scaler(
                        self,
                        sk_e.steps[1][1],
                        rasl_steps[1][1].impl,
                        msg=msg,
                    )


class TestMetrics(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.creditg = lale.datasets.openml.fetch(
            "credit-g", "classification", preprocess=True, astype="pandas"
        )

    def test_accuracy(self):
        (train_X, train_y), (test_X, test_y) = self.creditg
        est = LogisticRegression().fit(train_X, train_y)
        y_pred = est.predict(test_X)
        sk_score = sk_accuracy_score(test_y, y_pred)
        self.assertEqual(sk_score, rasl_accuracy_score(test_y, y_pred))
        rasl_scorer = rasl_get_scorer("accuracy")
        self.assertEqual(sk_score, rasl_scorer(est, test_X, test_y))
        batches = mockup_data_loader(test_X, test_y, 3, "pandas")
        self.assertEqual(sk_score, rasl_scorer.score_estimator_batched(est, batches))

    def test_r2_score(self):
        (train_X, train_y), (test_X, test_y) = self.creditg
        est = LogisticRegression().fit(train_X, train_y)
        y_pred = est.predict(test_X)
        sk_score = sk_r2_score(test_y, y_pred)
        self.assertAlmostEqual(sk_score, rasl_r2_score(test_y, y_pred))
        rasl_scorer = rasl_get_scorer("r2")
        self.assertAlmostEqual(sk_score, rasl_scorer(est, test_X, test_y))
        batches = mockup_data_loader(test_X, test_y, 3, "pandas")
        self.assertAlmostEqual(
            sk_score, rasl_scorer.score_estimator_batched(est, batches)
        )


class TestBatchedBaggingClassifier(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.creditg = lale.datasets.openml.fetch(
            "credit-g", "classification", preprocess=False, astype="pandas"
        )

    @classmethod
    def _make_sk_trainable(cls):
        from sklearn.tree import DecisionTreeClassifier as SkDecisionTreeClassifier

        return sk_make_pipeline(
            SkOrdinalEncoder(handle_unknown="use_encoded_value", unknown_value=-1),
            SkMinMaxScaler(),
            SkDecisionTreeClassifier(random_state=97, max_features="auto"),
        )

    @classmethod
    def _make_rasl_trainable(cls, final_est):
        if final_est == "bagging_monoid":
            est = BatchedBaggingClassifier(
                base_estimator=DecisionTreeClassifier(
                    random_state=97, max_features="auto"
                )
            )
        else:
            assert False, final_est
        return (
            RaslOrdinalEncoder(handle_unknown="use_encoded_value", unknown_value=-1)
            >> RaslMinMaxScaler()
            >> est
        )

    def test_classifier(self):
        (X_train, y_train), (X_test, y_test) = self.creditg
        import warnings

        clf = BatchedBaggingClassifier()
        # test_schemas_are_schemas
        lale.type_checking.validate_is_schema(clf.input_schema_fit())
        lale.type_checking.validate_is_schema(clf.input_schema_predict())
        lale.type_checking.validate_is_schema(clf.output_schema_predict())
        lale.type_checking.validate_is_schema(clf.hyperparam_schema())

        # test_init_fit_predict
        pipeline = self._make_rasl_trainable("bagging_monoid")

        trained = pipeline.fit(X_train, y_train)
        _ = trained.predict(X_test)

        # test_with_hyperopt
        from lale.lib.lale import Hyperopt

        (X_train, y_train), (X_test, _) = lale.datasets.openml.fetch(
            "credit-g", "classification", preprocess=True, astype="pandas"
        )

        hyperopt = Hyperopt(estimator=pipeline, max_evals=1, verbose=True)
        trained = hyperopt.fit(X_train, y_train)
        _ = trained.predict(X_test)
        # test_cross_validation
        from lale.helpers import cross_val_score

        cv_results = cross_val_score(pipeline, X_train, y_train, cv=2)
        self.assertEqual(len(cv_results), 2)

        # test_with_gridsearchcv_auto_wrapped
        from sklearn.metrics import accuracy_score, make_scorer

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from lale.lib.lale import GridSearchCV

            grid_search = GridSearchCV(
                estimator=pipeline,
                lale_num_samples=1,
                lale_num_grids=1,
                cv=2,
                scoring=make_scorer(accuracy_score),
            )
            grid_search.fit(X_train, y_train)

        # test_predict_on_trainable
        trained = clf.fit(X_train, y_train)
        clf.predict(X_train)

        # test_to_json
        clf.to_json()

    def test_fit_batching(self):
        (train_X, train_y), (test_X, test_y) = self.creditg
        from sklearn.metrics import accuracy_score

        train_data_space = train_X.memory_usage().sum() + train_y.memory_usage()
        unique_class_labels = list(train_y.unique())
        for n_batches in [1, 3]:
            for prio in [PrioStep(), PrioBatch()]:
                batches = mockup_data_loader(train_X, train_y, n_batches, "pandas")
                rasl_trainable = self._make_rasl_trainable("bagging_monoid")
                rasl_trained = fit_with_batches(
                    pipeline=rasl_trainable,
                    batches=batches,
                    scoring=None,
                    unique_class_labels=unique_class_labels,
                    max_resident=3 * math.ceil(train_data_space / n_batches),
                    prio=prio,
                    partial_transform=False,
                    verbose=0,
                    progress_callback=None,
                )
                predictions = rasl_trained.predict(test_X)
                rasl_acc = accuracy_score(test_y, predictions)
                if n_batches == 1:
                    sk_pipeline = self._make_sk_trainable()
                    sk_pipeline.fit(train_X, train_y)
                    predictions = sk_pipeline.predict(test_X)
                    sk_acc = accuracy_score(test_y, predictions)
                    self.assertEqual(rasl_acc, sk_acc)

    def test_cross_val_score(self):
        (train_X, train_y), (_, _) = self.creditg
        for n_batches in [1, 3]:
            _ = rasl_cross_val_score(
                pipeline=self._make_rasl_trainable("bagging_monoid"),
                batches=mockup_data_loader(train_X, train_y, n_batches, "pandas"),
                scoring=rasl_get_scorer("accuracy"),
                cv=KFold(3),
                unique_class_labels=list(train_y.unique()),
                max_resident=None,
                prio=PrioBatch(),
                same_fold=True,
                verbose=0,
            )



# Copyright 2021, 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
RASL operators and functions (experimental).

Relational Algebra Operators
============================

* lale.lib.rasl. `Aggregate`_
* lale.lib.rasl. `Alias`_
* lale.lib.rasl. `Filter`_
* lale.lib.rasl. `GroupBy`_
* lale.lib.rasl. `Join`_
* lale.lib.rasl. `Map`_
* lale.lib.rasl. `OrderBy`_
* lale.lib.rasl. `Project`_
* lale.lib.rasl. `Relational`_

Transformers
============

* lale.lib.rasl. `Batching`_
* lale.lib.rasl. `ConcatFeatures`_
* lale.lib.rasl. `Convert`_
* lale.lib.rasl. `Scan`_
* lale.lib.rasl. `SplitXy`_

Scikit-learn Operators
======================

* lale.lib.rasl. `MinMaxScaler`_
* lale.lib.rasl. `OneHotEncoder`_
* lale.lib.rasl. `OrdinalEncoder`_
* lale.lib.rasl. `HashingEncoder`_
* lale.lib.rasl. `SelectKBest`_
* lale.lib.rasl. `SimpleImputer`_
* lale.lib.rasl. `StandardScaler`_

Estimators
==========

* lale.lib.rasl. `BatchedBaggingClassifier`_

Functions
=========

* lale.lib.rasl. `categorical`_
* lale.lib.rasl. `date_time`_
* lale.lib.rasl. `SparkExplainer`_

Data Loaders
============
* lale.lib.rasl. `csv_data_loader`_
* lale.lib.rasl. `mockup_data_loader`_
* lale.lib.rasl. `openml_data_loader`_

Metrics
=======
* lale.lib.rasl. `accuracy_score`_
* lale.lib.rasl. `get_scorer`_
* lale.lib.rasl. `r2_score`_

Other Facilities
================

* lale.lib.rasl. `Prio`_
* lale.lib.rasl. `PrioBatch`_
* lale.lib.rasl. `PrioResourceAware`_
* lale.lib.rasl. `PrioStep`_
* lale.lib.rasl. `cross_val_score`_
* lale.lib.rasl. `cross_validate`_
* lale.lib.rasl. `fit_with_batches`_
* lale.lib.rasl. `is_associative`_
* lale.lib.rasl. `is_incremental`_

.. _`Aggregate`: lale.lib.rasl.aggregate.html
.. _`Alias`: lale.lib.rasl.alias.html
.. _`Filter`: lale.lib.rasl.filter.html
.. _`GroupBy`: lale.lib.rasl.group_by.html
.. _`Join`: lale.lib.rasl.join.html
.. _`Map`: lale.lib.rasl.map.html
.. _`OrderBy`: lale.lib.rasl.orderby.html
.. _`Project`: lale.lib.rasl.project.html
.. _`Relational`: lale.lib.rasl.relational.html

.. _`BatchedBaggingClassifier`: lale.lib.rasl.batched_bagging_classifier.html
.. _`Batching`: lale.lib.rasl.batching.html
.. _`ConcatFeatures`: lale.lib.rasl.concat_features.html
.. _`Convert`: lale.lib.rasl.convert.html
.. _`Scan`: lale.lib.rasl.scan.html
.. _`SplitXy`: lale.lib.rasl.split_xy.html

.. _`MinMaxScaler`: lale.lib.rasl.min_max_scaler.html
.. _`OneHotEncoder`: lale.lib.rasl.one_hot_encoder.html
.. _`OrdinalEncoder`: lale.lib.rasl.ordinal_encoder.html
.. _`HashingEncoder`: lale.lib.rasl.hashing_encoder.html
.. _`SelectKBest`: lale.lib.rasl.select_k_best.html
.. _`SimpleImputer`: lale.lib.rasl.simple_imputer.html
.. _`StandardScaler`: lale.lib.rasl.standard_scaler.html

.. _`BaggingMonoidClassifier`: lale.lib.rasl.batched_bagging_classifier.html

.. _`categorical`: lale.lib.rasl.functions.html#lale.lib.rasl.functions.categorical
.. _`date_time`: lale.lib.rasl.functions.html#lale.lib.rasl.functions.date_time
.. _`SparkExplainer`: lale.lib.rasl.spark_explainer.html

.. _`Prio`: lale.lib.rasl.Prio.html
.. _`PrioBatch`: lale.lib.rasl.PrioBatch.html
.. _`PrioResourceAware`: lale.lib.rasl.PrioResourceAware.html
.. _`PrioStep`: lale.lib.rasl.PrioStep.html
.. _`accuracy_score`: lale.lib.rasl.accuracy_score.html
.. _`cross_val_score`: lale.lib.rasl.cross_val_score.html
.. _`cross_validate`: lale.lib.rasl.cross_validate.html
.. _`fit_with_batches`: lale.lib.rasl.fit_with_batches.html
.. _`get_scorer`: lale.lib.rasl.get_scorer.html
.. _`is_associative`: lale.lib.rasl.is_associative.html
.. _`is_incremental`: lale.lib.rasl.is_incremental.html
.. _`csv_data_loader`: lale.lib.rasl.datasets.html#lale.lib.rasl.datasets.csv_data_loader
.. _`mockup_data_loader`: lale.lib.rasl.datasets.html#lale.lib.rasl.datasets.mockup_data_loader
.. _`openml_data_loader`: lale.lib.rasl.datasets.html#lale.lib.rasl.datasets.openml_data_loader
.. _`r2_score`: lale.lib.rasl.r2_score.html

"""

from ._task_graphs import (
    Prio,
    PrioBatch,
    PrioResourceAware,
    PrioStep,
    cross_val_score,
    cross_validate,
    fit_with_batches,
    is_associative,
    is_incremental,
)
from .aggregate import Aggregate
from .alias import Alias
from .batched_bagging_classifier import BatchedBaggingClassifier
from .batching import Batching
from .concat_features import ConcatFeatures
from .convert import Convert
from .datasets import csv_data_loader, mockup_data_loader, openml_data_loader
from .filter import Filter
from .functions import categorical, date_time
from .group_by import GroupBy
from .hashing_encoder import HashingEncoder
from .join import Join
from .map import Map
from .metrics import accuracy_score, get_scorer, r2_score
from .min_max_scaler import MinMaxScaler
from .monoid import Monoid, MonoidableOperator, MonoidFactory
from .one_hot_encoder import OneHotEncoder
from .orderby import OrderBy
from .ordinal_encoder import OrdinalEncoder
from .project import Project
from .relational import Relational
from .scan import Scan
from .select_k_best import SelectKBest
from .simple_imputer import SimpleImputer
from .spark_explainer import SparkExplainer
from .split_xy import SplitXy
from .standard_scaler import StandardScaler



# Copyright 2019-2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, List, Optional

import numpy as np
from typing_extensions import Protocol

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators
import lale.type_checking
from lale.expressions import it
from lale.lib.dataframe import column_index, get_columns
from lale.type_checking import is_schema, validate_is_schema

from .map import Map
from .monoid import Monoid, MonoidFactory


def _columns_schema_to_list(X, schema) -> List[column_index]:
    s_all = lale.datasets.data_schemas._to_schema(X)
    s_row = s_all["items"]
    n_columns = s_row["minItems"]
    assert n_columns == s_row["maxItems"]
    s_cols = s_row["items"]
    if isinstance(s_cols, dict):
        if lale.type_checking.is_subschema(s_cols, schema):
            result = get_columns(X)
        else:
            result = []
    else:
        assert isinstance(s_cols, list)
        cols = get_columns(X)
        result = [
            cols[i]
            for i in range(n_columns)
            if lale.type_checking.is_subschema(s_cols[i], schema)
        ]
    return result


class _ProjectMonoid(Monoid):
    def __init__(self, columns: Monoid, drop_columns: Monoid):
        self._columns = columns
        self._drop_columns = drop_columns

    def combine(self, other: "_ProjectMonoid"):
        c = self._columns.combine(other._columns)
        dc = self._drop_columns.combine(other._drop_columns)
        return _ProjectMonoid(c, dc)

    @property
    def is_absorbing(self):
        return self._columns.is_absorbing and self._drop_columns.is_absorbing


class _StaticMonoid(Monoid):
    def __init__(self, v):
        self._v = v

    def combine(self, other: "_StaticMonoid"):
        assert self._v == other._v
        return self

    @property
    def is_absorbing(self):
        return True


class _StaticMonoidFactory(MonoidFactory[Any, List[column_index], _StaticMonoid]):
    def __init__(self, cl):
        self._cl = cl

    def to_monoid(self, df):
        cl = self._cl
        if cl is None:
            cl = get_columns(df)
            self._cl = cl

        return _StaticMonoid(cl)

    def from_monoid(self, v: _StaticMonoid):
        return v._v


class _DynamicMonoidFactory(
    MonoidFactory[Any, List[column_index], _StaticMonoid], Protocol
):
    pass


class _CallableMonoidFactory(_DynamicMonoidFactory):
    def __init__(self, c):
        self._c = c

    def to_monoid(self, df):
        c = self._c
        if not isinstance(c, list):
            assert callable(c)
            c = c(df)
            self._c = c

        return _StaticMonoid(c)

    def from_monoid(self, v: _StaticMonoid):
        return v._v


class _SchemaMonoidFactory(_DynamicMonoidFactory):
    def __init__(self, c):
        self._c = c

    def to_monoid(self, df):
        c = self._c
        if not isinstance(c, list):
            assert isinstance(c, dict)
            c = _columns_schema_to_list(df, c)
            self._c = c

        return _StaticMonoid(c)

    def from_monoid(self, v: _StaticMonoid):
        return v._v


class _AllDataMonoidFactory(_CallableMonoidFactory):
    """This really needs all the data, and should not be considered a monoid.
    It is used to simplify the code, but does not enable monoidal fitting
    """

    def __init__(self, c):
        super().__init__(c)


def get_column_factory(columns, kind) -> MonoidFactory:
    if columns is None:
        if kind == "passthrough":
            return _StaticMonoidFactory(None)
        else:
            return _StaticMonoidFactory([])
    elif isinstance(columns, list):
        return _StaticMonoidFactory(columns)
    elif isinstance(columns, MonoidFactory):
        return columns
    elif callable(columns):
        return _AllDataMonoidFactory(columns)
    elif isinstance(columns, dict):
        validate_is_schema(columns)
        return _SchemaMonoidFactory(columns)
    else:
        raise TypeError(f"type {type(columns)}, columns {columns}")


class _ProjectImpl:
    def __init__(self, columns=None, drop_columns=None):
        self._columns = get_column_factory(columns, "passthrough")
        self._drop_columns = get_column_factory(drop_columns, "drop")
        self._monoid = None

        self._hyperparams = {"columns": columns, "drop_columns": drop_columns}

    def __getattribute__(self, item):
        # we want to remove fit if a static column is available
        # since it should be considered already trained
        if item == "fit":
            omit_fit = False
            try:
                cols = super().__getattribute__("_columns")
                drop_cols = super().__getattribute__("_drop_columns")
                if isinstance(cols, _StaticMonoidFactory) and isinstance(
                    drop_cols, _StaticMonoidFactory
                ):
                    omit_fit = True
            except AttributeError:
                pass
            if omit_fit:
                raise AttributeError(
                    "fit cannot be called on a Project that has a static columns and drop_columns or has already been fit"
                )
        elif item in ["to_monoid", "from_monoid", "partial_fit"]:
            omit_monoid = False
            try:
                cols = super().__getattribute__("_columns")
                drop_cols = super().__getattribute__("_drop_columns")
                if isinstance(cols, _AllDataMonoidFactory) or isinstance(
                    drop_cols, _AllDataMonoidFactory
                ):
                    omit_monoid = True
            except AttributeError:
                pass
            if omit_monoid:
                raise AttributeError("monoidal operations not available")
        return super().__getattribute__(item)

    def _to_monoid_internal(self, xy):
        df, _ = xy
        col = self._columns.to_monoid(df)
        dcol = self._drop_columns.to_monoid(df)

        return _ProjectMonoid(col, dcol)

    def to_monoid(self, xy):
        return self._to_monoid_internal(xy)

    def _from_monoid_internal(self, pm: _ProjectMonoid):
        col = self._columns.from_monoid(pm._columns)
        dcol = self._drop_columns.from_monoid(pm._drop_columns)

        self._fit_columns = [c for c in col if c not in dcol]

    def from_monoid(self, pm: _ProjectMonoid):
        self._from_monoid_internal(pm)

    _monoid: Optional[_ProjectMonoid]

    def partial_fit(self, X, y=None):
        if self._monoid is None or not self._monoid.is_absorbing:
            lifted = self._to_monoid_internal((X, y))
            if self._monoid is not None:  # not first fit
                lifted = self._monoid.combine(lifted)
            self._from_monoid_internal(lifted)
        return self

    def _fit_internal(self, X, y=None):
        lifted = self._to_monoid_internal((X, y))
        self._from_monoid_internal(lifted)
        return self

    def fit(self, X, y=None):
        return self._fit_internal(X, y)

    def transform(self, X):
        fitcols = getattr(self, "_fit_columns", None)
        if fitcols is None:
            self._fit_internal(X)
            fitcols = getattr(self, "_fit_columns", None)
            assert fitcols is not None

        if isinstance(X, np.ndarray):
            result = X[:, self._fit_columns]  # type: ignore
        else:
            # use the rasl backend
            cols = [
                X.columns[x] if isinstance(x, int) else x for x in self._fit_columns
            ]
            exprs = {c: it[c] for c in cols}
            m = Map(columns=exprs)
            assert isinstance(m, lale.operators.TrainedOperator)
            result = m.transform(X)
        # elif isinstance(X, pd.DataFrame):
        #     if len(self._fit_columns) == 0 or isinstance(self._fit_columns[0], int):
        #         result = X.iloc[:, self._fit_columns]
        #     else:
        #         result = X[self._fit_columns]
        # else:
        #     raise TypeError(f"type {type(X)}")
        s_X = lale.datasets.data_schemas._to_schema(X)
        s_result = self._transform_schema_nocheck(s_X)
        result = lale.datasets.data_schemas.add_schema(result, s_result, recalc=True)
        result = lale.datasets.data_schemas.add_table_name(
            result, lale.datasets.data_schemas.get_table_name(X)
        )
        return result

    def _transform_schema_nocheck(self, s_X):
        if hasattr(self, "_fit_columns"):
            return self._transform_schema_fit_columns(s_X)
        keep_cols = self._columns
        drop_cols = self._drop_columns
        known_keep_cols = False
        known_drop_cols = False
        if keep_cols is None:
            known_keep_cols = True
        elif isinstance(keep_cols, _SchemaMonoidFactory):
            kc = keep_cols._c
            keep_cols = kc
            known_keep_cols = True
        if drop_cols is None:
            known_drop_cols = True
        elif isinstance(drop_cols, _SchemaMonoidFactory):
            dc = drop_cols._c
            drop_cols = dc
            known_drop_cols = True
        if known_keep_cols and known_drop_cols:
            return self._transform_schema_schema(s_X, keep_cols, drop_cols)
        return s_X

    def transform_schema(self, s_X):
        """Used internally by Lale for type-checking downstream operators."""
        if is_schema(s_X):
            return self._transform_schema_nocheck(s_X)
        else:
            X = lale.datasets.data_schemas.add_schema(s_X)
            assert X is not None
            self.fit(X)
            return self._transform_schema_fit_columns(X.json_schema)

    def _transform_schema_fit_columns(self, s_X):
        s_X = lale.datasets.data_schemas._to_schema(s_X)
        s_row = s_X["items"]
        s_cols = s_row["items"]
        n_columns = len(self._fit_columns)
        if isinstance(s_cols, dict):
            s_cols_result = s_cols
        else:
            name2i = {s_cols[i]["description"]: i for i in range(len(s_cols))}
            keep_cols_i = [
                name2i[col] if isinstance(col, str) else col
                for col in self._fit_columns
            ]
            s_cols_result = [s_cols[i] for i in keep_cols_i]
        s_result = {
            **s_X,
            "items": {
                **s_row,
                "minItems": n_columns,
                "maxItems": n_columns,
                "items": s_cols_result,
            },
        }
        return s_result

    def _transform_schema_schema(self, s_X, s_keep, s_drop):
        def is_keeper(column_schema):
            if s_keep is not None:
                if not lale.type_checking.is_subschema(column_schema, s_keep):
                    return False
            if s_drop is not None:
                if lale.type_checking.is_subschema(column_schema, s_drop):
                    return False
            return True

        s_X = lale.datasets.data_schemas._to_schema(s_X)
        s_row = s_X["items"]
        s_cols = s_row["items"]
        if isinstance(s_cols, dict):
            if is_keeper(s_cols):
                s_row_result = s_row
            else:
                s_row_result = {"type": "array", "minItems": 0, "maxItems": 0}
        else:
            assert isinstance(s_cols, list)
            s_cols_result = [s for s in s_cols if is_keeper(s)]
            n_columns = len(s_cols_result)
            s_row_result = {
                "type": "array",
                "minItems": n_columns,
                "maxItems": n_columns,
                "items": s_cols_result,
            }
        return {"type": "array", "items": s_row_result}


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their types, one at a time, omitting cross-argument constraints, if any.",
            "type": "object",
            "additionalProperties": False,
            "required": ["columns", "drop_columns"],
            "relevantToOptimizer": [],
            "properties": {
                "columns": {
                    "description": """The subset of columns to retain.

The supported column specification formats include some of the ones
from scikit-learn's ColumnTransformer_, and in addition, filtering by
using a JSON subschema_ check.

.. _ColumnTransformer: https://scikit-learn.org/stable/modules/generated/sklearn.compose.ColumnTransformer.html
.. _subschema: https://github.com/IBM/jsonsubschema""",
                    "anyOf": [
                        {
                            "enum": [None],
                            "description": "If not specified, keep all columns.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "integer"},
                            "description": "Multiple columns by index.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "string"},
                            "description": "Multiple Dataframe columns by names.",
                        },
                        {
                            "laleType": "callable",
                            "description": "Callable that is passed the input data X and can return a list of column names or indices.",
                        },
                        {
                            "type": "object",
                            "description": "Keep columns whose schema is a subschema of this JSON schema.",
                        },
                    ],
                    "default": None,
                },
                "drop_columns": {
                    "description": """The subset of columns to remove.

The `drop_columns` argument supports the same formats as `columns`.
If both are specified, keep everything from `columns` that is not
also in `drop_columns`.""",
                    "anyOf": [
                        {
                            "enum": [None],
                            "description": "If not specified, drop no further columns.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "integer"},
                            "description": "Multiple columns by index.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "string"},
                            "description": "Multiple Dataframe columns by names.",
                        },
                        {
                            "laleType": "callable",
                            "description": "Callable that is passed the input data X and can return a list of column names or indices.",
                        },
                        {
                            "type": "object",
                            "description": "Remove columns whose schema is a subschema of this JSON schema.",
                        },
                    ],
                    "default": None,
                },
            },
        }
    ]
}

_input_fit_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        },
        "y": {"description": "Target for supervised learning (ignored)."},
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        }
    },
}

_output_transform_schema = {
    "description": "Features; the outer array is over samples.",
    "type": "array",
    "items": {
        "type": "array",
        "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
    },
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Projection keeps a subset of the columns, like in relational algebra.

Examples
--------
>>> df = pd.DataFrame(data={'A': [1,2], 'B': ['x','y'], 'C': [3,4]})
>>> keep_numbers = Project(columns={'type': 'number'})
>>> keep_numbers.fit(df).transform(df)
NDArrayWithSchema([[1, 3],
                   [2, 4]])
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.project.html",
    "import_from": "lale.lib.rasl",
    "type": "object",
    "tags": {"pre": ["categoricals"], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


Project = lale.operators.make_operator(_ProjectImpl, _combined_schemas)

lale.docstrings.set_docstrings(Project)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
import typing
from typing import Any, Tuple

import numpy as np
import pandas as pd

import lale.docstrings
import lale.operators
from lale.expressions import count, it, median, mode, replace, sum
from lale.helpers import _is_df, _is_pandas_df, _is_spark_df, _is_spark_with_index
from lale.lib.dataframe import get_columns
from lale.lib.sklearn import simple_imputer
from lale.schemas import Enum

from .aggregate import Aggregate
from .map import Map
from .monoid import Monoid, MonoidableOperator


def _is_numeric_df(X):
    if _is_pandas_df(X):
        return X.shape[1] == X.select_dtypes(include=np.number).shape[1]
    elif _is_spark_df(X):
        from pyspark.sql.types import NumericType

        numeric_cols = [
            f.name for f in X.schema.fields if isinstance(f.dataType, NumericType)
        ]
        if _is_spark_with_index(X):
            for index_name in X.index_names:
                if index_name in numeric_cols:
                    numeric_cols.remove(index_name)
        return len(get_columns(X)) == len(numeric_cols)
    else:
        return False


def _is_string_df(X):
    if _is_pandas_df(X):
        return X.shape[1] == X.select_dtypes(include="object").shape[1]
    elif _is_spark_df(X):
        from pyspark.sql.types import StringType

        numeric_cols = [
            f.name for f in X.schema.fields if isinstance(f.dataType, StringType)
        ]
        return len(get_columns(X)) == len(numeric_cols)
    else:
        return False


class _SimpleImputerMonoid(Monoid):
    def __init__(self, *, feature_names_in_, lifted_statistics, strategy):
        self.feature_names_in_ = feature_names_in_
        self.lifted_statistics = lifted_statistics
        self.strategy = strategy

    def combine(self, other: "_SimpleImputerMonoid"):
        assert list(self.feature_names_in_) == list(other.feature_names_in_)
        if self.strategy == "constant":
            assert self.lifted_statistics.equals(other.lifted_statistics)
            combined_statistic = self.lifted_statistics
        elif self.strategy == "mean":
            combined_statistic = {}
            combined_statistic["sum"] = (
                self.lifted_statistics["sum"] + other.lifted_statistics["sum"]
            )
            combined_statistic["count"] = (
                self.lifted_statistics["count"] + other.lifted_statistics["count"]
            )
        else:
            raise ValueError(
                "_combine is only supported for imputation strategy `mean` and `constant`."
            )
        return _SimpleImputerMonoid(
            feature_names_in_=self.feature_names_in_,
            lifted_statistics=combined_statistic,
            strategy=self.strategy,
        )


class _SimpleImputerImpl(MonoidableOperator[_SimpleImputerMonoid]):
    def __init__(
        self,
        missing_values=np.nan,
        strategy="mean",
        fill_value=None,
        verbose=0,
        copy=True,
        add_indicator=False,
    ):
        self._hyperparams = {}
        self._hyperparams["missing_values"] = missing_values
        self._hyperparams["strategy"] = strategy
        self._hyperparams["fill_value"] = fill_value
        self._hyperparams["verbose"] = verbose
        if not copy:
            raise ValueError("This implementation only supports `copy=True`.")
        self._hyperparams["copy"] = copy
        if add_indicator:
            raise ValueError("This implementation only supports `add_indicator=False`.")
        self._hyperparams["add_indicator"] = add_indicator
        # the `indicator_`` property is always None as we do not support `add_indicator=True`
        self.indicator_ = None

    def to_monoid(self, v: Tuple[Any, Any]):
        hyperparams = self._hyperparams
        X, _ = v
        feature_names_in_ = get_columns(X)
        agg_data = None
        # learn the values to be imputed
        strategy = hyperparams["strategy"]
        if strategy == "constant":
            fill_value = _SimpleImputerImpl._get_fill_value(X, hyperparams)
            agg_data = [[fill_value for col in get_columns(X)]]
            lifted_statistics = pd.DataFrame(agg_data, columns=get_columns(X))
        elif strategy == "mean":
            agg_op_sum = Aggregate(
                columns={c: sum(it[c]) for c in get_columns(X)},
                exclude_value=hyperparams["missing_values"],
            )
            agg_op_count = Aggregate(
                columns={c: count(it[c]) for c in get_columns(X)},
                exclude_value=hyperparams["missing_values"],
            )
            lifted_statistics = {}
            agg_sum = agg_op_sum.transform(X)
            if agg_sum is not None and _is_spark_df(agg_sum):
                agg_sum = agg_sum.toPandas()
            agg_count = agg_op_count.transform(X)
            if agg_count is not None and _is_spark_df(agg_count):
                agg_count = agg_count.toPandas()
            lifted_statistics["sum"] = agg_sum
            lifted_statistics["count"] = agg_count
        else:
            raise ValueError(
                "SimpleImputer can create a Monoind only for imputation strategy `mean` and `constant`."
            )
        return _SimpleImputerMonoid(
            feature_names_in_=feature_names_in_,
            lifted_statistics=lifted_statistics,
            strategy=strategy,
        )

    def from_monoid(self, lifted):
        self._monoid = lifted
        self.feature_names_in_ = lifted.feature_names_in_
        self.n_features_in_ = len(self.feature_names_in_)
        _lifted_statistics = lifted.lifted_statistics
        strategy = self._hyperparams["strategy"]
        if strategy == "constant":
            self.statistics_ = _lifted_statistics.to_numpy()[0]
        elif strategy == "mean":
            self.statistics_ = (
                _lifted_statistics["sum"] / _lifted_statistics["count"]
            ).to_numpy()[0]
        self._transformer = None

    def fit(self, X, y=None):

        self._validate_input(X)

        agg_op = None
        agg_data = None
        # learn the values to be imputed
        if self._hyperparams["strategy"] in ["mean", "constant"]:
            lifted = self.to_monoid((X, y))
            self.from_monoid(lifted)
            return self
        elif self._hyperparams["strategy"] == "median":
            agg_op = Aggregate(
                columns={c: median(it[c]) for c in get_columns(X)},
                exclude_value=self._hyperparams["missing_values"],
            )
        elif self._hyperparams["strategy"] == "most_frequent":
            agg_op = Aggregate(
                columns={c: mode(it[c]) for c in get_columns(X)},
                exclude_value=self._hyperparams["missing_values"],
            )
        if agg_data is None and agg_op is not None:
            agg_data = agg_op.transform(X)
        self.feature_names_in_ = get_columns(X)
        self.n_features_in_ = len(self.feature_names_in_)
        if agg_data is not None and _is_spark_df(agg_data):
            agg_data = agg_data.toPandas()
        if agg_data is not None and _is_pandas_df(agg_data):
            self.statistics_ = agg_data.to_numpy()[
                0
            ]  # Converting from a 2-d array to 1-d
        self._transformer = None
        return self

    def _build_transformer(self):
        # prepare the transformer
        transformer = Map(
            columns={
                col_name: replace(
                    it[col_name],
                    {self._hyperparams["missing_values"]: self.statistics_[col_idx]},
                )
                for col_idx, col_name in enumerate(self.feature_names_in_)
            }
        )
        return transformer

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer()
        return self._transformer.transform(X)

    @staticmethod
    def _get_fill_value(X, hyperparams):
        # assign appropriate value to fill_value depending on the datatype.
        # default fill_value is 0 for numerical input and "missing_value"
        # otherwise
        if hyperparams["fill_value"] is None:
            if _is_numeric_df(X):
                fill_value = 0
            else:
                fill_value = "missing_value"
        else:
            fill_value = hyperparams["fill_value"]
        # validate that fill_value is numerical for numerical data
        if (
            hyperparams["strategy"] == "constant"
            and _is_numeric_df(X)
            and not isinstance(fill_value, numbers.Real)
        ):
            raise ValueError(
                "'fill_value'={0} is invalid. Expected a "
                "numerical value when imputing numerical "
                "data".format(fill_value)
            )
        return fill_value

    def _validate_input(self, X):
        # validate that the dataset is either a pandas dataframe or spark.
        # For example, sparse matrix is not allowed.
        if not _is_df(X):
            raise ValueError(
                f"""Unsupported type(X) {type(X)} for SimpleImputer.
            Only pandas.DataFrame or pyspark.sql.DataFrame are allowed."""
            )
        # validate input to check the correct dtype and strategy
        # `mean` and `median` are not applicable to string inputs
        if not _is_numeric_df(X) and self._hyperparams["strategy"] in [
            "mean",
            "median",
        ]:
            raise ValueError(
                "Cannot use {} strategy with non-numeric data.".format(
                    self._hyperparams["strategy"]
                )
            )

        # Check that missing_values are the right type
        if _is_numeric_df(X) and not isinstance(
            self._hyperparams["missing_values"], numbers.Real
        ):
            raise ValueError(
                "'X' and 'missing_values' types are expected to be"
                " both numerical. Got X.dtypes={} and "
                " type(missing_values)={}.".format(
                    X.dtypes, type(self._hyperparams["missing_values"])
                )
            )


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn's `SimpleImputer`_.
Works on both pandas and Spark dataframes by using `Aggregate`_ for `fit` and `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`SimpleImputer`: https://scikit-learn.org/stable/modules/generated/sklearn.imputer.SimpleImputer.html
.. _`Aggregate`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.aggregate.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.simple_imputer.html",
    "type": "object",
    "tags": {
        "pre": [],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": simple_imputer._hyperparams_schema,
        "input_fit": simple_imputer._input_fit_schema,
        "input_transform": simple_imputer._input_transform_schema,
        "output_transform": simple_imputer._output_transform_schema,
    },
}

SimpleImputer = lale.operators.make_operator(_SimpleImputerImpl, _combined_schemas)

SimpleImputer = typing.cast(
    lale.operators.PlannedIndividualOp,
    SimpleImputer.customize_schema(
        copy=Enum(
            values=[True],
            desc="`copy=True` is the only value currently supported by this implementation",
            default=True,
        ),
        add_indicator=Enum(
            values=[False],
            desc="`add_indicator=False` is the only value currently supported by this implementation",
            default=False,
        ),
    ),
)

lale.docstrings.set_docstrings(SimpleImputer)
